"""Test zha sensor."""

import asyncio
from collections.abc import Awaitable, Callable
from datetime import UTC, datetime
from functools import partial
import math
from typing import Any
from unittest.mock import AsyncMock, MagicMock, patch

import pytest
from zhaquirks.danfoss import thermostat as danfoss_thermostat
from zigpy.device import Device as ZigpyDevice
from zigpy.profiles import zha
import zigpy.profiles.zha
from zigpy.quirks import CustomCluster, DeviceRegistry, get_device
from zigpy.quirks.v2 import CustomDeviceV2, QuirkBuilder, ReportingConfig
from zigpy.quirks.v2.homeassistant import EntityPlatform, EntityType, UnitOfMass
from zigpy.quirks.v2.homeassistant.sensor import (
    SensorDeviceClass as SensorDeviceClassV2,
)
import zigpy.types as t
from zigpy.zcl import Cluster
from zigpy.zcl.clusters import general, homeautomation, hvac, measurement, smartenergy
from zigpy.zcl.clusters.general import AnalogInput, PowerConfiguration
from zigpy.zcl.clusters.general_const import AnalogInputType, ApplicationType
from zigpy.zcl.clusters.manufacturer_specific import ManufacturerSpecificCluster

from tests.common import (
    SIG_EP_INPUT,
    SIG_EP_OUTPUT,
    SIG_EP_PROFILE,
    SIG_EP_TYPE,
    create_mock_zigpy_device,
    get_entity,
    join_zigpy_device,
    send_attributes_report,
    zigpy_device_from_json,
)
from zha.application import Platform
from zha.application.const import ZCL_INIT_ATTRS, ZHA_CLUSTER_HANDLER_READS_PER_REQ
from zha.application.gateway import Gateway
from zha.application.platforms import PlatformEntity, sensor
from zha.application.platforms.sensor import (
    AnalogInputSensor,
    DanfossSoftwareErrorCode,
    Temperature,
)
from zha.application.platforms.sensor.const import SensorDeviceClass, SensorStateClass
from zha.application.platforms.sensor.helpers import resolution_to_decimal_precision
from zha.units import (
    PERCENTAGE,
    UnitOfElectricPotential,
    UnitOfEnergy,
    UnitOfPressure,
    UnitOfVolume,
)
from zha.zigbee.cluster_handlers import AttrReportConfig
from zha.zigbee.cluster_handlers.manufacturerspecific import OppleRemoteClusterHandler
from zha.zigbee.device import Device

EMAttrs = homeautomation.ElectricalMeasurement.AttributeDefs


def elec_measurement_zigpy_device_mock(
    zha_gateway: Gateway,
) -> ZigpyDevice:
    """Electric Measurement zigpy device."""

    zigpy_device = create_mock_zigpy_device(
        zha_gateway,
        {
            1: {
                SIG_EP_INPUT: [
                    general.Basic.cluster_id,
                    homeautomation.ElectricalMeasurement.cluster_id,
                ],
                SIG_EP_OUTPUT: [],
                SIG_EP_TYPE: zigpy.profiles.zha.DeviceType.SIMPLE_SENSOR,
                SIG_EP_PROFILE: zigpy.profiles.zha.PROFILE_ID,
            }
        },
    )
    zigpy_device.node_desc.mac_capability_flags |= 0b_0000_0100
    zigpy_device.endpoints[1].electrical_measurement.PLUGGED_ATTR_READS = {
        "ac_current_divisor": 10,
        "ac_current_multiplier": 1,
        "ac_power_divisor": 10,
        "ac_power_multiplier": 1,
        "ac_voltage_divisor": 10,
        "ac_voltage_multiplier": 1,
        "measurement_type": 0x48,  # PHASE_A_MEASUREMENT | DC_MEASUREMENT
        "power_divisor": 10,
        "power_multiplier": 1,
        "dc_voltage_divisor": 10,
        "dc_voltage_multiplier": 1,
        "dc_current_divisor": 10,
        "dc_current_multiplier": 1,
        "dc_power_divisor": 10,
        "dc_power_multiplier": 1,
    }
    return zigpy_device


def metering_zigpy_device_mock(
    zha_gateway: Gateway,
) -> ZigpyDevice:
    """Metering zigpy device."""

    zigpy_device = create_mock_zigpy_device(
        zha_gateway,
        {
            1: {
                SIG_EP_INPUT: [
                    general.Basic.cluster_id,
                    smartenergy.Metering.cluster_id,
                ],
                SIG_EP_OUTPUT: [],
                SIG_EP_TYPE: zigpy.profiles.zha.DeviceType.SIMPLE_SENSOR,
                SIG_EP_PROFILE: zigpy.profiles.zha.PROFILE_ID,
            }
        },
        model="TS011F",  # polling is enabled for this model
    )
    zigpy_device.node_desc.mac_capability_flags |= 0b_0000_0100
    zigpy_device.endpoints[1].smartenergy_metering.PLUGGED_ATTR_READS = {
        "unit_of_measure": 0,  # kWh
        "multiplier": 1,
        "divisor": 1000,
        "summation_formatting": 0b0_0100_011,  # read from plug
        "metering_device_type": 0,  # electric
    }
    return zigpy_device


async def async_test_humidity(
    zha_gateway: Gateway, cluster: Cluster, entity: PlatformEntity
) -> None:
    """Test humidity sensor."""
    await send_attributes_report(zha_gateway, cluster, {1: 1, 0: 1000, 2: 100})
    assert_state(entity, 10.0, "%")


async def async_test_flow(
    zha_gateway: Gateway, cluster: Cluster, entity: PlatformEntity
) -> None:
    """Test flow sensor."""
    await send_attributes_report(zha_gateway, cluster, {1: 1, 0: 40})
    assert_state(entity, 4.0, "m³/h")

    await send_attributes_report(zha_gateway, cluster, {1: 1, 0: 0xFFFF})
    assert_state(entity, None, "m³/h")


async def async_test_temperature(
    zha_gateway: Gateway, cluster: Cluster, entity: PlatformEntity
) -> None:
    """Test temperature sensor."""
    assert entity.extra_state_attribute_names is None
    await send_attributes_report(zha_gateway, cluster, {1: 1, 0: 2900, 2: 100})
    assert_state(entity, 29.0, "°C")


async def async_test_pressure(
    zha_gateway: Gateway, cluster: Cluster, entity: PlatformEntity
) -> None:
    """Test pressure sensor."""
    await send_attributes_report(zha_gateway, cluster, {1: 1, 0: 1000, 2: 10000})
    assert_state(entity, 1000, "hPa")

    await send_attributes_report(zha_gateway, cluster, {0: 1000, 20: -1, 16: 10000})
    assert_state(entity, 1000, "hPa")


async def async_test_illuminance(
    zha_gateway: Gateway, cluster: Cluster, entity: PlatformEntity
) -> None:
    """Test illuminance sensor."""
    await send_attributes_report(zha_gateway, cluster, {1: 1, 0: 10, 2: 20})
    assert_state(entity, 1.0, "lx")

    await send_attributes_report(zha_gateway, cluster, {0: 0xFFFF})
    assert_state(entity, None, "lx")


async def async_test_metering(
    zha_gateway: Gateway, cluster: Cluster, entity: PlatformEntity
) -> None:
    """Test Smart Energy metering sensor."""
    assert entity.extra_state_attribute_names == {
        "status",
        "device_type",
        "zcl_unit_of_measurement",
    }
    await send_attributes_report(
        zha_gateway, cluster, {1025: 1, 1024: 12345, 1026: 100}
    )
    assert_state(entity, 12345.0, None)
    assert entity.state["status"] == "NO_ALARMS"
    assert entity.state["device_type"] == "Electric Metering"

    await send_attributes_report(zha_gateway, cluster, {1024: 12346, "status": 64 + 8})
    assert_state(entity, 12346.0, None)
    assert entity.state["status"] in (
        "SERVICE_DISCONNECT|POWER_FAILURE",
        "POWER_FAILURE|SERVICE_DISCONNECT",
    )

    await send_attributes_report(
        zha_gateway, cluster, {"status": 64 + 8, "metering_device_type": 1}
    )
    assert entity.state["status"] in (
        "SERVICE_DISCONNECT|NOT_DEFINED",
        "NOT_DEFINED|SERVICE_DISCONNECT",
    )

    await send_attributes_report(
        zha_gateway, cluster, {"status": 64 + 8, "metering_device_type": 2}
    )
    assert entity.state["status"] in (
        "SERVICE_DISCONNECT|PIPE_EMPTY",
        "PIPE_EMPTY|SERVICE_DISCONNECT",
    )

    await send_attributes_report(
        zha_gateway, cluster, {"status": 64 + 8, "metering_device_type": 5}
    )
    assert entity.state["status"] in (
        "SERVICE_DISCONNECT|TEMPERATURE_SENSOR",
        "TEMPERATURE_SENSOR|SERVICE_DISCONNECT",
    )

    # Status for other meter types
    await send_attributes_report(
        zha_gateway, cluster, {"status": 32, "metering_device_type": 4}
    )
    assert entity.state["status"] in ("<bitmap8.32: 32>", "32")


async def async_test_smart_energy_summation_delivered(
    zha_gateway: Gateway, cluster, entity
):
    """Test SmartEnergy Summation delivered sensor."""
    assert entity.extra_state_attribute_names == {
        "status",
        "device_type",
        "zcl_unit_of_measurement",
    }
    await send_attributes_report(
        zha_gateway, cluster, {1025: 1, "current_summ_delivered": 12321, 1026: 100}
    )
    assert_state(entity, 12.321, UnitOfEnergy.KILO_WATT_HOUR)
    assert entity.state["status"] == "NO_ALARMS"
    assert entity.state["device_type"] == "Electric Metering"
    assert entity.info_object.device_class == SensorDeviceClass.ENERGY


async def async_test_smart_energy_summation_received(
    zha_gateway: Gateway, cluster, entity
):
    """Test SmartEnergy Summation received sensor."""

    await send_attributes_report(
        zha_gateway, cluster, {1025: 1, "current_summ_received": 12321, 1026: 100}
    )
    assert_state(entity, 12.321, UnitOfEnergy.KILO_WATT_HOUR)
    assert entity.state["status"] == "NO_ALARMS"
    assert entity.state["device_type"] == "Electric Metering"
    assert entity.info_object.device_class == SensorDeviceClass.ENERGY


async def async_test_smart_energy_summation(
    zha_gateway: Gateway, cluster: Cluster, entity: PlatformEntity
) -> None:
    """Test SmartEnergy Summation delivered sensro."""

    await send_attributes_report(
        zha_gateway, cluster, {1025: 1, "current_summ_delivered": 12321, 1026: 100}
    )
    assert_state(entity, 12.32, "m³")
    assert entity.state["status"] == "NO_ALARMS"
    assert entity.state["device_type"] == "Electric Metering"


async def async_test_electrical_measurement(
    zha_gateway: Gateway, cluster: Cluster, entity: PlatformEntity
) -> None:
    """Test electrical measurement sensor."""
    # update divisor cached value
    await send_attributes_report(zha_gateway, cluster, {"ac_power_divisor": 1})
    await send_attributes_report(
        zha_gateway, cluster, {0: 1, EMAttrs.active_power.id: 100}
    )
    assert_state(entity, 100, "W")

    await send_attributes_report(
        zha_gateway, cluster, {0: 1, EMAttrs.active_power.id: 99}
    )
    assert_state(entity, 99, "W")

    await send_attributes_report(zha_gateway, cluster, {"ac_power_divisor": 10})
    await send_attributes_report(
        zha_gateway, cluster, {0: 1, EMAttrs.active_power.id: 1000}
    )
    assert_state(entity, 100, "W")

    await send_attributes_report(
        zha_gateway, cluster, {0: 1, EMAttrs.active_power.id: 99}
    )
    assert_state(entity, 9.9, "W")

    await send_attributes_report(zha_gateway, cluster, {0: 1, 0x050D: 88})
    assert entity.state["active_power_max"] == 8.8


async def async_test_em_apparent_power(
    zha_gateway: Gateway, cluster: Cluster, entity: PlatformEntity
) -> None:
    """Test electrical measurement Apparent Power sensor."""
    # update divisor cached value
    await send_attributes_report(zha_gateway, cluster, {"ac_power_divisor": 1})
    await send_attributes_report(zha_gateway, cluster, {0: 1, 0x050F: 100})
    assert_state(entity, 100, "VA")

    await send_attributes_report(zha_gateway, cluster, {0: 1, 0x050F: 99})
    assert_state(entity, 99, "VA")

    await send_attributes_report(zha_gateway, cluster, {"ac_power_divisor": 10})
    await send_attributes_report(zha_gateway, cluster, {0: 1, 0x050F: 1000})
    assert_state(entity, 100, "VA")

    await send_attributes_report(zha_gateway, cluster, {0: 1, 0x050F: 99})
    assert_state(entity, 9.9, "VA")


async def async_test_em_power_factor(
    zha_gateway: Gateway, cluster: Cluster, entity: PlatformEntity
):
    """Test electrical measurement Power Factor sensor."""
    assert entity.extra_state_attribute_names == {"measurement_type"}

    # update divisor cached value
    await send_attributes_report(zha_gateway, cluster, {"ac_power_divisor": 1})
    await send_attributes_report(zha_gateway, cluster, {0: 1, 0x0510: 100, 10: 1000})
    assert_state(entity, 100, PERCENTAGE)

    await send_attributes_report(zha_gateway, cluster, {0: 1, 0x0510: 99, 10: 1000})
    assert_state(entity, 99, PERCENTAGE)

    await send_attributes_report(zha_gateway, cluster, {"ac_power_divisor": 10})
    await send_attributes_report(zha_gateway, cluster, {0: 1, 0x0510: 100, 10: 5000})
    assert_state(entity, 100, PERCENTAGE)

    await send_attributes_report(zha_gateway, cluster, {0: 1, 0x0510: 99, 10: 5000})
    assert_state(entity, 99, PERCENTAGE)


async def async_test_em_rms_current(
    current_attrid: int,
    current_max_attrid: int,
    current_max_attr_name: str,
    zha_gateway: Gateway,
    cluster: Cluster,
    entity: PlatformEntity,
) -> None:
    """Test electrical measurement RMS Current sensor."""

    await send_attributes_report(zha_gateway, cluster, {0: 1, current_attrid: 1234})
    assert_state(entity, 1.234, "A")

    await send_attributes_report(zha_gateway, cluster, {"ac_current_divisor": 10})
    await send_attributes_report(zha_gateway, cluster, {0: 1, current_attrid: 236})
    assert_state(entity, 23.6, "A")

    await send_attributes_report(zha_gateway, cluster, {0: 1, current_attrid: 1236})
    assert_state(entity, 123.6, "A")

    await send_attributes_report(zha_gateway, cluster, {0: 1, current_max_attrid: 88})
    assert entity.state[current_max_attr_name] == 8.8


async def async_test_em_rms_voltage(
    zha_gateway: Gateway, cluster: Cluster, entity: PlatformEntity
) -> None:
    """Test electrical measurement RMS Voltage sensor."""
    assert entity.extra_state_attribute_names == {"measurement_type", "rms_voltage_max"}

    await send_attributes_report(zha_gateway, cluster, {0: 1, 0x0505: 1234})
    assert_state(entity, 123.4, "V")

    await send_attributes_report(zha_gateway, cluster, {0: 1, 0x0505: 234})
    assert_state(entity, 23.4, "V")

    await send_attributes_report(zha_gateway, cluster, {"ac_voltage_divisor": 100})
    await send_attributes_report(zha_gateway, cluster, {0: 1, 0x0505: 2236})
    assert_state(entity, 22.36, "V")

    await send_attributes_report(zha_gateway, cluster, {0: 1, 0x0507: 888})
    assert entity.state["rms_voltage_max"] == 8.88


async def async_test_powerconfiguration(
    zha_gateway: Gateway, cluster: Cluster, entity: PlatformEntity
) -> None:
    """Test powerconfiguration/battery sensor."""
    assert entity.extra_state_attribute_names == {
        "battery_voltage",
        "battery_quantity",
        "battery_size",
    }
    await send_attributes_report(zha_gateway, cluster, {33: 98})
    assert_state(entity, 49, "%")
    assert entity.state["battery_voltage"] == 2.9
    assert entity.state["battery_quantity"] == 3
    assert entity.state["battery_size"] == "AAA"
    await send_attributes_report(zha_gateway, cluster, {32: 20})
    assert entity.state["battery_voltage"] == 2.0


async def async_test_powerconfiguration2(
    zha_gateway: Gateway, cluster: Cluster, entity: PlatformEntity
):
    """Test powerconfiguration/battery sensor."""
    await send_attributes_report(zha_gateway, cluster, {33: 255})
    assert_state(entity, None, "%")

    await send_attributes_report(zha_gateway, cluster, {33: 98})
    assert_state(entity, 49, "%")


async def async_test_device_temperature(
    zha_gateway: Gateway, cluster: Cluster, entity: PlatformEntity
) -> None:
    """Test temperature sensor."""
    await send_attributes_report(zha_gateway, cluster, {0: 2900})
    assert_state(entity, 29.0, "°C")


async def async_test_setpoint_change_source(
    zha_gateway: Gateway, cluster: Cluster, entity: PlatformEntity
):
    """Test the translation of numerical state into enum text."""
    await send_attributes_report(
        zha_gateway,
        cluster,
        {hvac.Thermostat.AttributeDefs.setpoint_change_source.id: 0x01},
    )
    assert entity.state["state"] == "Schedule"


async def async_test_pi_heating_demand(
    zha_gateway: Gateway, cluster: Cluster, entity: PlatformEntity
):
    """Test pi heating demand is correctly returned."""
    await send_attributes_report(
        zha_gateway, cluster, {hvac.Thermostat.AttributeDefs.pi_heating_demand.id: 1}
    )
    assert_state(entity, 1, "%")


async def async_test_change_source_timestamp(
    zha_gateway: Gateway, cluster: Cluster, entity: PlatformEntity
):
    """Test change source timestamp is correctly returned."""
    await send_attributes_report(
        zha_gateway,
        cluster,
        {hvac.Thermostat.AttributeDefs.setpoint_change_source_timestamp.id: 781355715},
    )
    assert entity.state["state"] == datetime(2024, 10, 4, 11, 15, 15, tzinfo=UTC)


async def async_test_em_dc_voltage(
    zha_gateway: Gateway, cluster: Cluster, entity: PlatformEntity
) -> None:
    """Test electrical measurement DC Voltage sensor."""
    assert entity.extra_state_attribute_names == {"measurement_type"}

    await send_attributes_report(zha_gateway, cluster, {0: 1, 0x0100: 1234})
    assert_state(entity, 123.4, "V")

    await send_attributes_report(zha_gateway, cluster, {0: 1, 0x0100: 234})
    assert_state(entity, 23.4, "V")

    await send_attributes_report(zha_gateway, cluster, {"dc_voltage_divisor": 100})
    await send_attributes_report(zha_gateway, cluster, {0: 1, 0x0100: 2236})
    assert_state(entity, 22.36, "V")


async def async_test_em_dc_current(
    zha_gateway: Gateway, cluster: Cluster, entity: PlatformEntity
) -> None:
    """Test electrical measurement DC Current sensor."""

    await send_attributes_report(zha_gateway, cluster, {0: 1, 0x0103: 1234})
    assert_state(entity, 1.234, "A")

    await send_attributes_report(zha_gateway, cluster, {"dc_current_divisor": 10})
    await send_attributes_report(zha_gateway, cluster, {0: 1, 0x0103: 236})
    assert_state(entity, 23.6, "A")

    await send_attributes_report(zha_gateway, cluster, {0: 1, 0x0103: 1236})
    assert_state(entity, 123.6, "A")


async def async_test_em_dc_power(
    zha_gateway: Gateway, cluster: Cluster, entity: PlatformEntity
) -> None:
    """Test electrical measurement DC Power sensor."""
    # update divisor cached value
    await send_attributes_report(zha_gateway, cluster, {"dc_power_divisor": 1})
    await send_attributes_report(zha_gateway, cluster, {0: 1, 0x0106: 100})
    assert_state(entity, 100, "W")

    await send_attributes_report(zha_gateway, cluster, {0: 1, 0x0106: 99})
    assert_state(entity, 99, "W")

    await send_attributes_report(zha_gateway, cluster, {"dc_power_divisor": 10})
    await send_attributes_report(zha_gateway, cluster, {0: 1, 0x0106: 1000})
    assert_state(entity, 100, "W")

    await send_attributes_report(zha_gateway, cluster, {0: 1, 0x0106: 99})
    assert_state(entity, 9.9, "W")


@pytest.mark.parametrize(
    "cluster_id, entity_type, test_func, read_plug, unsupported_attrs",
    (
        (
            measurement.RelativeHumidity.cluster_id,
            sensor.Humidity,
            async_test_humidity,
            None,
            None,
        ),
        (
            measurement.FlowMeasurement.cluster_id,
            sensor.Flow,
            async_test_flow,
            None,
            None,
        ),
        (
            measurement.TemperatureMeasurement.cluster_id,
            sensor.Temperature,
            async_test_temperature,
            None,
            None,
        ),
        (
            measurement.PressureMeasurement.cluster_id,
            sensor.Pressure,
            async_test_pressure,
            None,
            None,
        ),
        (
            measurement.IlluminanceMeasurement.cluster_id,
            sensor.Illuminance,
            async_test_illuminance,
            None,
            None,
        ),
        (
            smartenergy.Metering.cluster_id,
            sensor.SmartEnergyMetering,
            async_test_metering,
            {
                "demand_formatting": 0xF9,
                "divisor": 1,
                "metering_device_type": 0x00,
                "multiplier": 1,
                "status": 0x00,
            },
            {"current_summ_delivered"},
        ),
        (
            smartenergy.Metering.cluster_id,
            sensor.SmartEnergySummation,
            async_test_smart_energy_summation,
            {
                "demand_formatting": 0xF9,
                "divisor": 1000,
                "metering_device_type": 0x00,
                "multiplier": 1,
                "status": 0x00,
                "summation_formatting": 0b1_0111_010,
                "unit_of_measure": 0x01,
            },
            {"instantaneous_demand"},
        ),
        (
            smartenergy.Metering.cluster_id,
            sensor.SmartEnergySummationReceived,
            async_test_smart_energy_summation_received,
            {
                "demand_formatting": 0xF9,
                "divisor": 1000,
                "metering_device_type": 0x00,
                "multiplier": 1,
                "status": 0x00,
                "summation_formatting": 0b1_0111_010,
                "unit_of_measure": 0x00,
                "current_summ_received": 0,
            },
            {"instantaneous_demand", "current_summ_delivered"},
        ),
        (
            homeautomation.ElectricalMeasurement.cluster_id,
            sensor.PolledElectricalMeasurement,
            async_test_electrical_measurement,
            {"ac_power_divisor": 1000, "ac_power_multiplier": 1},
            {"apparent_power", "rms_current", "rms_voltage"},
        ),
        (
            homeautomation.ElectricalMeasurement.cluster_id,
            sensor.ElectricalMeasurementApparentPower,
            async_test_em_apparent_power,
            {"ac_power_divisor": 1000, "ac_power_multiplier": 1},
            {"active_power", "rms_current", "rms_voltage"},
        ),
        (
            homeautomation.ElectricalMeasurement.cluster_id,
            sensor.ElectricalMeasurementPowerFactor,
            async_test_em_power_factor,
            {"ac_power_divisor": 1000, "ac_power_multiplier": 1},
            {"active_power", "apparent_power", "rms_current", "rms_voltage"},
        ),
        (
            homeautomation.ElectricalMeasurement.cluster_id,
            sensor.ElectricalMeasurementRMSCurrent,
            partial(async_test_em_rms_current, 0x0508, 0x050A, "rms_current_max"),
            {"ac_current_divisor": 1000, "ac_current_multiplier": 1},
            {"active_power", "apparent_power", "rms_voltage"},
        ),
        (
            homeautomation.ElectricalMeasurement.cluster_id,
            sensor.ElectricalMeasurementRMSCurrentPhB,
            partial(async_test_em_rms_current, 0x0908, 0x090A, "rms_current_max_ph_b"),
            {
                "ac_current_divisor": 1000,
                "ac_current_multiplier": 1,
                "rms_current_ph_b": 0,
            },
            {"active_power", "apparent_power", "rms_voltage"},
        ),
        (
            homeautomation.ElectricalMeasurement.cluster_id,
            sensor.ElectricalMeasurementRMSCurrentPhC,
            partial(async_test_em_rms_current, 0x0A08, 0x0A0A, "rms_current_max_ph_c"),
            {
                "ac_current_divisor": 1000,
                "ac_current_multiplier": 1,
                "rms_current_ph_c": 0,
            },
            {"active_power", "apparent_power", "rms_voltage"},
        ),
        (
            homeautomation.ElectricalMeasurement.cluster_id,
            sensor.ElectricalMeasurementRMSVoltage,
            async_test_em_rms_voltage,
            {"ac_voltage_divisor": 10, "ac_voltage_multiplier": 1},
            {"active_power", "apparent_power", "rms_current"},
        ),
        (
            general.PowerConfiguration.cluster_id,
            sensor.Battery,
            async_test_powerconfiguration,
            {
                "battery_size": 4,  # AAA
                "battery_voltage": 29,
                "battery_quantity": 3,
            },
            None,
        ),
        (
            general.PowerConfiguration.cluster_id,
            sensor.Battery,
            async_test_powerconfiguration2,
            {
                "battery_size": 4,  # AAA
                "battery_voltage": 29,
                "battery_quantity": 3,
            },
            None,
        ),
        (
            general.DeviceTemperature.cluster_id,
            sensor.DeviceTemperature,
            async_test_device_temperature,
            None,
            None,
        ),
        (
            hvac.Thermostat.cluster_id,
            sensor.SetpointChangeSource,
            async_test_setpoint_change_source,
            None,
            None,
        ),
        (
            hvac.Thermostat.cluster_id,
            sensor.PiHeatingDemand,
            async_test_pi_heating_demand,
            None,
            None,
        ),
        (
            hvac.Thermostat.cluster_id,
            sensor.SetpointChangeSourceTimestamp,
            async_test_change_source_timestamp,
            None,
            None,
        ),
        (
            homeautomation.ElectricalMeasurement.cluster_id,
            sensor.ElectricalMeasurementDCVoltage,
            async_test_em_dc_voltage,
            {"dc_voltage_divisor": 10, "dc_voltage_multiplier": 1, "dc_voltage": 0},
            {"active_power", "apparent_power", "rms_current", "rms_voltage"},
        ),
        (
            homeautomation.ElectricalMeasurement.cluster_id,
            sensor.ElectricalMeasurementDCCurrent,
            async_test_em_dc_current,
            {"dc_current_divisor": 1000, "dc_current_multiplier": 1, "dc_current": 0},
            {"active_power", "apparent_power", "rms_current", "rms_voltage"},
        ),
        (
            homeautomation.ElectricalMeasurement.cluster_id,
            sensor.ElectricalMeasurementDCPower,
            async_test_em_dc_power,
            {"dc_power_divisor": 1000, "dc_power_multiplier": 1, "dc_power": 0},
            {"active_power", "apparent_power", "rms_current", "rms_voltage"},
        ),
    ),
)
async def test_sensor(
    zha_gateway: Gateway,
    cluster_id: int,
    entity_type: type[PlatformEntity],
    test_func: Callable[[Cluster, Cluster, PlatformEntity], Awaitable[None]],
    read_plug: dict | None,
    unsupported_attrs: set | None,
) -> None:
    """Test zha sensor platform."""

    zigpy_device = create_mock_zigpy_device(
        zha_gateway,
        {
            1: {
                SIG_EP_INPUT: [cluster_id, general.Basic.cluster_id],
                SIG_EP_OUTPUT: [],
                SIG_EP_TYPE: zigpy.profiles.zha.DeviceType.ON_OFF_SWITCH,
                SIG_EP_PROFILE: zigpy.profiles.zha.PROFILE_ID,
            }
        },
    )
    cluster = zigpy_device.endpoints[1].in_clusters[cluster_id]
    if unsupported_attrs:
        for attr in unsupported_attrs:
            cluster.add_unsupported_attribute(attr)
    if cluster_id in (
        smartenergy.Metering.cluster_id,
        homeautomation.ElectricalMeasurement.cluster_id,
    ):
        # this one is mains powered
        zigpy_device.node_desc.mac_capability_flags |= 0b_0000_0100
    cluster.PLUGGED_ATTR_READS = read_plug or {}

    zha_device = await join_zigpy_device(zha_gateway, zigpy_device)
    entity = get_entity(
        zha_device, platform=Platform.SENSOR, exact_entity_type=entity_type
    )

    await zha_gateway.async_block_till_done()

    # test sensor associated logic
    await test_func(zha_gateway, cluster, entity)
    assert entity.fallback_name is None


async def test_analog_input_simple(zha_gateway: Gateway) -> None:
    """Test analog input sensors."""
    zigpy_dev = await zigpy_device_from_json(
        zha_gateway.application_controller,
        "tests/data/devices/isilentllc-masterbed-light-controller.json",
    )

    # Pretend this device has a proper description for the attribute
    # TODO: replace this unit test with one that uses a real device
    zigpy_dev.endpoints[2].analog_input.update_attribute(
        AnalogInput.AttributeDefs.description.id, "Some description"
    )

    zha_dev = await join_zigpy_device(zha_gateway, zigpy_dev)
    entity = get_entity(
        zha_dev, platform=Platform.SENSOR, exact_entity_type=AnalogInputSensor
    )

    assert entity.state["available"] is True
    assert entity.state["state"] == 2.1322579383850098
    assert entity.info_object.fallback_name == "Some description"
    assert entity.info_object.translation_key is None
    assert entity.info_object.unit == UnitOfElectricPotential.VOLT
    assert entity.info_object.device_class is None
    assert entity.info_object.suggested_display_precision is None


async def test_analog_input_ignored(zha_gateway: Gateway) -> None:
    """Test analog input sensors."""
    zigpy_dev = await zigpy_device_from_json(
        zha_gateway.application_controller,
        "tests/data/devices/isilentllc-masterbed-light-controller.json",
    )

    # Pretend this device has a proper description for the attribute
    # TODO: replace this unit test with one that uses a real device
    zigpy_dev.endpoints[2].analog_input.update_attribute(
        AnalogInput.AttributeDefs.description.id, "Some description"
    )

    # No units or application type are defined, however
    zigpy_dev.endpoints[2].analog_input.update_attribute(
        AnalogInput.AttributeDefs.engineering_units.id, 5
    )
    zigpy_dev.endpoints[2].analog_input.update_attribute(
        AnalogInput.AttributeDefs.engineering_units.id, None
    )
    zigpy_dev.endpoints[2].analog_input.add_unsupported_attribute(
        AnalogInput.AttributeDefs.engineering_units.id
    )
    # Also remove from PLUGGED_ATTR_READS so read_attributes doesn't restore the value
    zigpy_dev.endpoints[2].analog_input.PLUGGED_ATTR_READS.pop(
        AnalogInput.AttributeDefs.engineering_units.id, None
    )

    zha_dev = await join_zigpy_device(zha_gateway, zigpy_dev)

    # The entity is not created
    with pytest.raises(KeyError):
        get_entity(
            zha_dev, platform=Platform.SENSOR, exact_entity_type=AnalogInputSensor
        )


async def test_analog_input_complex(zha_gateway: Gateway) -> None:
    """Test analog input sensors."""
    zigpy_dev = await zigpy_device_from_json(
        zha_gateway.application_controller,
        "tests/data/devices/isilentllc-masterbed-light-controller.json",
    )

    # Pretend this device has a proper description for the attribute
    # TODO: replace this unit test with one that uses a real device
    zigpy_dev.endpoints[2].analog_input.update_attribute(
        AnalogInput.AttributeDefs.description.id, "Some description"
    )
    zigpy_dev.endpoints[2].analog_input.update_attribute(
        AnalogInput.AttributeDefs.application_type.id,
        int(
            ApplicationType(
                group=0, type=AnalogInputType.Relative_Humidity_Percent, index=0
            )
        ),
    )
    zigpy_dev.endpoints[2].analog_input.update_attribute(
        AnalogInput.AttributeDefs.resolution.id, 0.01
    )

    zha_dev = await join_zigpy_device(zha_gateway, zigpy_dev)
    entity = get_entity(
        zha_dev, platform=Platform.SENSOR, exact_entity_type=AnalogInputSensor
    )

    assert entity.state["available"] is True
    assert entity.state["state"] == 2.1322579383850098
    assert entity.info_object.fallback_name == "Some description"
    assert entity.info_object.translation_key is None
    assert entity.info_object.unit is PERCENTAGE  # overridden!
    assert entity.info_object.device_class is SensorDeviceClass.HUMIDITY  # overridden!
    assert entity.info_object.suggested_display_precision == 2


def assert_state(entity: PlatformEntity, state: Any, unit_of_measurement: str) -> None:
    """Check that the state is what is expected.

    This is used to ensure that the logic in each sensor class handled the
    attribute report it received correctly.
    """
    assert entity.state["state"] == state
    assert entity.info_object.unit == unit_of_measurement


async def test_electrical_measurement_init(
    zha_gateway: Gateway,
    caplog: pytest.LogCaptureFixture,
) -> None:
    """Test proper initialization of the electrical measurement cluster."""

    cluster_id = homeautomation.ElectricalMeasurement.cluster_id
    zigpy_device = create_mock_zigpy_device(
        zha_gateway,
        {
            1: {
                SIG_EP_INPUT: [cluster_id, general.Basic.cluster_id],
                SIG_EP_OUTPUT: [],
                SIG_EP_TYPE: zigpy.profiles.zha.DeviceType.ON_OFF_SWITCH,
                SIG_EP_PROFILE: zigpy.profiles.zha.PROFILE_ID,
            }
        },
    )

    cluster = zigpy_device.endpoints[1].in_clusters[cluster_id]
    zha_device = await join_zigpy_device(zha_gateway, zigpy_device)

    entity = get_entity(
        zha_device,
        platform=Platform.SENSOR,
        exact_entity_type=sensor.PolledElectricalMeasurement,
    )

    await send_attributes_report(
        zha_gateway,
        cluster,
        {EMAttrs.active_power.id: 100},
    )
    assert entity.state["state"] == 100

    cluster_handler = list(zha_device._endpoints.values())[0].all_cluster_handlers[
        "1:0x0b04"
    ]
    assert cluster_handler.ac_power_divisor == 1
    assert cluster_handler.ac_power_multiplier == 1

    # update power divisor
    await send_attributes_report(
        zha_gateway,
        cluster,
        {EMAttrs.active_power.id: 20, EMAttrs.power_divisor.id: 5},
    )
    assert cluster_handler.ac_power_divisor == 5
    assert cluster_handler.ac_power_multiplier == 1
    assert entity.state["state"] == 4.0

    zha_device.on_network = False

    await asyncio.sleep(entity.__polling_interval + 1)
    await zha_gateway.async_block_till_done(wait_background_tasks=True)
    assert (
        "-1-2820: skipping polling for updated state, available: False, allow polled requests: True"
        in caplog.text
    )

    zha_device.on_network = True

    await send_attributes_report(
        zha_gateway,
        cluster,
        {EMAttrs.active_power.id: 30, EMAttrs.ac_power_divisor.id: 10},
    )
    assert cluster_handler.ac_power_divisor == 10
    assert cluster_handler.ac_power_multiplier == 1
    assert entity.state["state"] == 3.0

    # update power multiplier
    await send_attributes_report(
        zha_gateway,
        cluster,
        {EMAttrs.active_power.id: 20, EMAttrs.power_multiplier.id: 6},
    )
    assert cluster_handler.ac_power_divisor == 10
    assert cluster_handler.ac_power_multiplier == 6
    assert entity.state["state"] == 12.0

    await send_attributes_report(
        zha_gateway,
        cluster,
        {EMAttrs.active_power.id: 30, EMAttrs.ac_power_multiplier.id: 20},
    )
    assert cluster_handler.ac_power_divisor == 10
    assert cluster_handler.ac_power_multiplier == 20
    assert entity.state["state"] == 60.0

    entity._refresh = AsyncMock(wraps=entity._refresh)

    assert entity._refresh.await_count == 0

    entity.disable()

    assert entity.enabled is False

    await asyncio.sleep(entity.__polling_interval + 1)
    await zha_gateway.async_block_till_done(wait_background_tasks=True)

    assert entity._refresh.await_count == 0

    entity.enable()

    assert entity.enabled is True

    await asyncio.sleep(entity.__polling_interval + 1)
    await zha_gateway.async_block_till_done(wait_background_tasks=True)

    assert entity._refresh.await_count == 1


@pytest.mark.parametrize(
    (
        "cluster_id",
        "unsupported_attributes",
        "included_entity_types",
        "excluded_entity_types",
    ),
    (
        (
            homeautomation.ElectricalMeasurement.cluster_id,
            {
                "apparent_power",
                "rms_voltage",
                "rms_current",
            },
            {
                sensor.PolledElectricalMeasurement,
                sensor.ElectricalMeasurementFrequency,
                sensor.ElectricalMeasurementPowerFactor,
            },
            {
                sensor.ElectricalMeasurementApparentPower,
                sensor.ElectricalMeasurementRMSVoltage,
                sensor.ElectricalMeasurementRMSCurrent,
            },
        ),
        (
            homeautomation.ElectricalMeasurement.cluster_id,
            {
                "apparent_power",
                "rms_current",
                "ac_frequency",
                "power_factor",
            },
            {
                sensor.ElectricalMeasurementRMSVoltage,
                sensor.PolledElectricalMeasurement,
            },
            {
                sensor.ElectricalMeasurementApparentPower,
                sensor.ElectricalMeasurementRMSCurrent,
                sensor.ElectricalMeasurementFrequency,
                sensor.ElectricalMeasurementPowerFactor,
            },
        ),
        (
            homeautomation.ElectricalMeasurement.cluster_id,
            set(),
            {
                sensor.ElectricalMeasurementRMSVoltage,
                sensor.PolledElectricalMeasurement,
                sensor.ElectricalMeasurementApparentPower,
                sensor.ElectricalMeasurementRMSCurrent,
                sensor.ElectricalMeasurementFrequency,
                sensor.ElectricalMeasurementPowerFactor,
            },
            set(),
        ),
        (
            smartenergy.Metering.cluster_id,
            {
                "instantaneous_demand",
            },
            {
                sensor.SmartEnergySummation,
            },
            {
                sensor.SmartEnergyMetering,
            },
        ),
        (
            smartenergy.Metering.cluster_id,
            {
                "instantaneous_demand",
                "current_summ_delivered",
            },
            set(),
            {
                sensor.SmartEnergyMetering,
                sensor.SmartEnergySummation,
            },
        ),
        (
            smartenergy.Metering.cluster_id,
            set(),
            {
                sensor.SmartEnergyMetering,
                sensor.SmartEnergySummation,
            },
            set(),
        ),
    ),
)
async def test_unsupported_attributes_sensor(
    zha_gateway: Gateway,
    cluster_id: int,
    unsupported_attributes: set,
    included_entity_types: set,
    excluded_entity_types: set,
) -> None:
    """Test zha sensor platform."""

    zigpy_device = create_mock_zigpy_device(
        zha_gateway,
        {
            1: {
                SIG_EP_INPUT: [cluster_id, general.Basic.cluster_id],
                SIG_EP_OUTPUT: [],
                SIG_EP_TYPE: zigpy.profiles.zha.DeviceType.ON_OFF_SWITCH,
                SIG_EP_PROFILE: zigpy.profiles.zha.PROFILE_ID,
            }
        },
    )
    cluster = zigpy_device.endpoints[1].in_clusters[cluster_id]
    if cluster_id == smartenergy.Metering.cluster_id:
        # this one is mains powered
        zigpy_device.node_desc.mac_capability_flags |= 0b_0000_0100
    for attr in unsupported_attributes:
        cluster.add_unsupported_attribute(attr)

    zha_device = await join_zigpy_device(zha_gateway, zigpy_device)

    present_entity_types = {
        type(e)
        for e in zha_device.platform_entities.values()
        if e.PLATFORM == Platform.SENSOR
        and ("lqi" not in e.unique_id and "rssi" not in e.unique_id)
    }

    assert included_entity_types <= present_entity_types
    assert not excluded_entity_types & present_entity_types


@pytest.mark.parametrize(
    "raw_uom, raw_value, expected_state, expected_uom",
    (
        (
            1,
            12320,
            1.23,
            UnitOfVolume.CUBIC_METERS,
        ),
        (
            1,
            1232000,
            123.2000,
            UnitOfVolume.CUBIC_METERS,
        ),
        (
            3,
            2340,
            0.23,
            UnitOfVolume.CUBIC_FEET,
        ),
        (
            3,
            2360,
            0.24,
            UnitOfVolume.CUBIC_FEET,
        ),
        (
            8,
            23660,
            2.37,
            UnitOfPressure.KPA,
        ),
        (
            0,
            9366,
            0.9366,
            UnitOfEnergy.KILO_WATT_HOUR,
        ),
        (
            0,
            999,
            0.0999,
            UnitOfEnergy.KILO_WATT_HOUR,
        ),
        (
            0,
            10091,
            1.0091,
            UnitOfEnergy.KILO_WATT_HOUR,
        ),
        (
            0,
            10099,
            1.0099,
            UnitOfEnergy.KILO_WATT_HOUR,
        ),
        (
            0,
            100999,
            10.0999,
            UnitOfEnergy.KILO_WATT_HOUR,
        ),
        (
            0,
            100023,
            10.0023,
            UnitOfEnergy.KILO_WATT_HOUR,
        ),
        (
            0,
            102456,
            10.2456,
            UnitOfEnergy.KILO_WATT_HOUR,
        ),
        (
            5,
            102456,
            10.25,
            "IMP gal",
        ),
        (
            7,
            50124,
            5.01,
            UnitOfVolume.LITERS,
        ),
    ),
)
async def test_se_summation_uom(
    zha_gateway: Gateway,
    raw_uom: int,
    raw_value: int,
    expected_state: str,
    expected_uom: str,
) -> None:
    """Test zha smart energy summation."""

    zigpy_device = create_mock_zigpy_device(
        zha_gateway,
        {
            1: {
                SIG_EP_INPUT: [
                    smartenergy.Metering.cluster_id,
                    general.Basic.cluster_id,
                ],
                SIG_EP_OUTPUT: [],
                SIG_EP_TYPE: zigpy.profiles.zha.DeviceType.SIMPLE_SENSOR,
                SIG_EP_PROFILE: zigpy.profiles.zha.PROFILE_ID,
            }
        },
    )
    zigpy_device.node_desc.mac_capability_flags |= 0b_0000_0100

    cluster = zigpy_device.endpoints[1].in_clusters[smartenergy.Metering.cluster_id]
    for attr in ("instantaneous_demand",):
        cluster.add_unsupported_attribute(attr)
    cluster.PLUGGED_ATTR_READS = {
        "current_summ_delivered": raw_value,
        "demand_formatting": 0xF9,
        "divisor": 10000,
        "metering_device_type": 0x00,
        "multiplier": 1,
        "status": 0x00,
        "summation_formatting": 0b1_0111_010,
        "unit_of_measure": raw_uom,
    }
    zha_device = await join_zigpy_device(zha_gateway, zigpy_device)

    entity = get_entity(
        zha_device, platform=Platform.SENSOR, qualifier="summation_delivered"
    )

    assert_state(entity, expected_state, expected_uom)


@pytest.mark.parametrize(
    "raw_measurement_type, expected_type",
    (
        (1, "ACTIVE_MEASUREMENT"),
        (8, "PHASE_A_MEASUREMENT"),
        (9, "ACTIVE_MEASUREMENT, PHASE_A_MEASUREMENT"),
        (
            15,
            "ACTIVE_MEASUREMENT, REACTIVE_MEASUREMENT, APPARENT_MEASUREMENT, PHASE_A_MEASUREMENT",
        ),
    ),
)
async def test_elec_measurement_sensor_type(
    raw_measurement_type: int,
    expected_type: str,
    zha_gateway: Gateway,
) -> None:
    """Test zha electrical measurement sensor type."""

    zigpy_dev = elec_measurement_zigpy_device_mock(zha_gateway)
    zigpy_dev.endpoints[1].electrical_measurement.PLUGGED_ATTR_READS[
        "measurement_type"
    ] = raw_measurement_type

    zha_dev = await join_zigpy_device(zha_gateway, zigpy_dev)

    entity = get_entity(
        zha_dev,
        platform=Platform.SENSOR,
        entity_type=sensor.ElectricalMeasurementApparentPower,
    )
    assert entity.state["measurement_type"] == expected_type


async def test_elec_measurement_sensor_polling(zha_gateway: Gateway) -> None:
    """Test ZHA electrical measurement sensor polling."""

    zigpy_dev = elec_measurement_zigpy_device_mock(zha_gateway)
    zigpy_dev.endpoints[1].electrical_measurement.PLUGGED_ATTR_READS["active_power"] = (
        20
    )

    zha_dev = await join_zigpy_device(zha_gateway, zigpy_dev)

    # test that the sensor has an initial state of 2.0
    entity = get_entity(
        zha_dev,
        platform=Platform.SENSOR,
        exact_entity_type=sensor.PolledElectricalMeasurement,
    )
    assert entity.state["state"] == 2.0

    # update the value for the power reading
    zigpy_dev.endpoints[1].electrical_measurement.PLUGGED_ATTR_READS["active_power"] = (
        60
    )

    # ensure the state is still 2.0
    assert entity.state["state"] == 2.0

    # let the polling happen
    await asyncio.sleep(90)
    await zha_gateway.async_block_till_done(wait_background_tasks=True)

    # ensure the state has been updated to 6.0
    assert entity.state["state"] == 6.0


async def test_metering_sensor_polling(zha_gateway: Gateway) -> None:
    """Test ZHA metering sensor polling."""

    zigpy_dev = metering_zigpy_device_mock(zha_gateway)
    zigpy_dev.endpoints[1].smartenergy_metering.PLUGGED_ATTR_READS[
        "current_summ_delivered"
    ] = 2000

    zha_dev = await join_zigpy_device(zha_gateway, zigpy_dev)

    # test that the sensor has an initial state of 2.0
    entity = get_entity(
        zha_dev,
        platform=Platform.SENSOR,
        exact_entity_type=sensor.PolledSmartEnergySummation,
    )
    assert entity.state["state"] == 2.0

    # update the value for the power reading
    zigpy_dev.endpoints[1].smartenergy_metering.PLUGGED_ATTR_READS[
        "current_summ_delivered"
    ] = 6000

    # ensure the state is still 2.0
    assert entity.state["state"] == 2.0

    # let the polling happen
    await asyncio.sleep(90)
    await zha_gateway.async_block_till_done(wait_background_tasks=True)

    # ensure the state has been updated to 6.0
    assert entity.state["state"] == 6.0


@pytest.mark.parametrize(
    "supported_attributes",
    (
        set(),
        {
            "active_power",
            "active_power_max",
            "rms_current",
            "rms_current_max",
            "rms_voltage",
            "rms_voltage_max",
        },
        {
            "active_power",
        },
        {
            "active_power",
            "active_power_max",
        },
        {
            "rms_current",
            "rms_current_max",
        },
        {
            "rms_voltage",
            "rms_voltage_max",
        },
    ),
)
async def test_elec_measurement_skip_unsupported_attribute(
    zha_gateway: Gateway,
    supported_attributes: set[str],
) -> None:
    """Test zha electrical measurement skipping update of unsupported attributes."""

    elec_measurement_zigpy_dev = elec_measurement_zigpy_device_mock(zha_gateway)
    zha_dev = await join_zigpy_device(zha_gateway, elec_measurement_zigpy_dev)

    cluster = zha_dev.device.endpoints[1].electrical_measurement

    all_attrs = {
        "active_power",
        "active_power_ph_b",
        "active_power_ph_c",
        "active_power_max",
        "active_power_max_ph_b",
        "active_power_max_ph_c",
        "total_active_power",
        "apparent_power",
        "rms_current",
        "rms_current_ph_b",
        "rms_current_ph_c",
        "rms_current_max",
        "rms_current_max_ph_b",
        "rms_current_max_ph_c",
        "rms_voltage",
        "rms_voltage_ph_b",
        "rms_voltage_ph_c",
        "rms_voltage_max",
        "rms_voltage_max_ph_b",
        "rms_voltage_max_ph_c",
        "power_factor",
        "power_factor_ph_b",
        "power_factor_ph_c",
        "ac_frequency",
        "ac_frequency_max",
        "ac_voltage_divisor",
        "ac_current_divisor",
        "ac_power_divisor",
        "ac_voltage_multiplier",
        "ac_power_multiplier",
        "ac_current_multiplier",
        "power_divisor",
        "power_multiplier",
        "dc_voltage",
        "dc_voltage_divisor",
        "dc_voltage_multiplier",
        "dc_current",
        "dc_current_divisor",
        "dc_current_multiplier",
        "dc_power",
        "dc_power_divisor",
        "dc_power_multiplier",
    }
    for attr in all_attrs - supported_attributes:
        cluster.add_unsupported_attribute(attr)
    cluster.read_attributes.reset_mock()

    entity = get_entity(
        zha_dev,
        platform=Platform.SENSOR,
        exact_entity_type=sensor.PolledElectricalMeasurement,
    )
    await entity.async_update()
    await zha_dev.gateway.async_block_till_done()
    assert cluster.read_attributes.call_count == math.ceil(
        len(supported_attributes) / ZHA_CLUSTER_HANDLER_READS_PER_REQ
    )
    read_attrs = {
        a for call in cluster.read_attributes.call_args_list for a in call[0][0]
    }
    assert read_attrs == supported_attributes


class TimestampCluster(CustomCluster, ManufacturerSpecificCluster):
    """Timestamp Quirk V2 Cluster."""

    cluster_id = 0xEF00
    ep_attribute = "time_test_cluster"
    attributes = {
        0xEF65: ("start_time", t.uint32_t, True),
    }

    def __init__(self, *args, **kwargs) -> None:
        """Initialize."""
        super().__init__(*args, **kwargs)
        # populate cache to create config entity
        self._attr_cache.update({0xEF65: 10})


(
    QuirkBuilder("Fake_Timestamp_sensor", "Fake_Model_sensor")
    .replaces(TimestampCluster)
    .sensor(
        "start_time",
        TimestampCluster.cluster_id,
        device_class=SensorDeviceClassV2.TIMESTAMP,  # Use the zigpy enum
        attribute_converter=lambda x: datetime.fromtimestamp(x + 946684800, tz=UTC),
        translation_key="start_time",
        fallback_name="Start Time",
    )
    .add_to_registry()
)


async def zigpy_device_timestamp_sensor_v2_mock(
    zha_gateway: Gateway,  # pylint: disable=unused-argument
):
    """Timestamp Test device."""

    zigpy_device = create_mock_zigpy_device(
        zha_gateway,
        {
            1: {
                SIG_EP_INPUT: [
                    general.Basic.cluster_id,
                    TimestampCluster.cluster_id,
                ],
                SIG_EP_OUTPUT: [],
                SIG_EP_TYPE: zigpy.profiles.zha.DeviceType.ON_OFF_SWITCH,
                SIG_EP_PROFILE: zigpy.profiles.zha.PROFILE_ID,
            }
        },
        manufacturer="Fake_Timestamp_sensor",
        model="Fake_Model_sensor",
    )

    zigpy_device = get_device(zigpy_device)

    zha_device = await join_zigpy_device(zha_gateway, zigpy_device)
    return zha_device, zigpy_device.endpoints[1].time_test_cluster


async def test_timestamp_sensor_v2(zha_gateway: Gateway) -> None:
    """Test quirks defined sensor."""

    zha_device, cluster = await zigpy_device_timestamp_sensor_v2_mock(zha_gateway)
    assert isinstance(zha_device.device, CustomDeviceV2)
    entity = get_entity(zha_device, platform=Platform.SENSOR, qualifier="start_time")

    await send_attributes_report(zha_gateway, cluster, {0xEF65: 781355715})
    assert entity.state["state"] == datetime(2024, 10, 4, 11, 15, 15, tzinfo=UTC)


class OppleCluster(CustomCluster, ManufacturerSpecificCluster):
    """Aqara manufacturer specific cluster."""

    cluster_id = 0xFCC0
    ep_attribute = "opple_cluster"
    attributes = {
        0x010C: ("last_feeding_size", t.uint16_t, True),
    }

    def __init__(self, *args, **kwargs) -> None:
        """Initialize."""
        super().__init__(*args, **kwargs)
        # populate cache to create config entity
        self._attr_cache.update({0x010C: 10})


(
    QuirkBuilder("Fake_Manufacturer_sensor", "Fake_Model_sensor")
    .replaces(OppleCluster)
    .sensor(
        "last_feeding_size",
        OppleCluster.cluster_id,
        divisor=1,
        multiplier=1,
        unit=UnitOfMass.GRAMS,
        translation_key="last_feeding_size",
        fallback_name="Last feeding size",
        reporting_config=ReportingConfig(
            min_interval=0, max_interval=60, reportable_change=1
        ),
    )
    .sensor(
        "power",
        OppleCluster.cluster_id,
        state_class=SensorStateClass.MEASUREMENT,
        device_class=SensorDeviceClass.POWER,
        fallback_name="Measurement",
    )
    .sensor(
        "energy",
        OppleCluster.cluster_id,
        state_class=SensorStateClass.TOTAL,
        device_class=SensorDeviceClass.ENERGY,
        fallback_name="Measurement",
    )
    .sensor(
        "energy_delivered",
        OppleCluster.cluster_id,
        state_class=SensorStateClass.TOTAL_INCREASING,
        device_class=SensorDeviceClass.ENERGY,
        fallback_name="Measurement",
    )
    .sensor(
        "energy_invalid_state_class",
        OppleCluster.cluster_id,
        state_class=SensorDeviceClass.ENERGY,
        device_class=SensorDeviceClass.ENERGY,
        fallback_name="Measurement",
    )
    .add_to_registry()
)


async def zigpy_device_aqara_sensor_v2_mock(
    zha_gateway: Gateway,  # pylint: disable=unused-argument
):
    """Device tracker zigpy Aqara motion sensor device."""

    zigpy_device = create_mock_zigpy_device(
        zha_gateway,
        {
            1: {
                SIG_EP_INPUT: [
                    general.Basic.cluster_id,
                    OppleCluster.cluster_id,
                ],
                SIG_EP_OUTPUT: [],
                SIG_EP_TYPE: zigpy.profiles.zha.DeviceType.OCCUPANCY_SENSOR,
                SIG_EP_PROFILE: zigpy.profiles.zha.PROFILE_ID,
            }
        },
        manufacturer="Fake_Manufacturer_sensor",
        model="Fake_Model_sensor",
    )

    zigpy_device = get_device(zigpy_device)

    zha_device = await join_zigpy_device(zha_gateway, zigpy_device)
    return zha_device, zigpy_device.endpoints[1].opple_cluster


async def test_last_feeding_size_sensor_v2(zha_gateway: Gateway) -> None:
    """Test quirks defined sensor."""

    zha_device, cluster = await zigpy_device_aqara_sensor_v2_mock(zha_gateway)
    assert isinstance(zha_device.device, CustomDeviceV2)
    entity = get_entity(
        zha_device, platform=Platform.SENSOR, qualifier="last_feeding_size"
    )

    await send_attributes_report(zha_gateway, cluster, {0x010C: 1})
    assert_state(entity, 1.0, "g")

    await send_attributes_report(zha_gateway, cluster, {0x010C: 5})
    assert_state(entity, 5.0, "g")


async def test_state_class(
    zha_gateway: Gateway, caplog: pytest.LogCaptureFixture
) -> None:
    """Test quirks defined sensor."""

    zha_device, cluster = await zigpy_device_aqara_sensor_v2_mock(zha_gateway)
    assert isinstance(zha_device.device, CustomDeviceV2)
    power_entity = get_entity(
        zha_device,
        platform=Platform.SENSOR,
        qualifier_func=lambda e: e.info_object.unique_id.endswith("power"),
    )
    energy_entity = get_entity(
        zha_device,
        platform=Platform.SENSOR,
        qualifier_func=lambda e: e.info_object.unique_id.endswith("energy"),
    )
    energy_delivered_entity = get_entity(
        zha_device, platform=Platform.SENSOR, qualifier="energy_delivered"
    )
    energy_invalid_state_class = get_entity(
        zha_device, platform=Platform.SENSOR, qualifier="energy_invalid_state_class"
    )

    assert power_entity.state_class == SensorStateClass.MEASUREMENT
    assert energy_entity.state_class == SensorStateClass.TOTAL
    assert energy_delivered_entity.state_class == SensorStateClass.TOTAL_INCREASING
    assert energy_invalid_state_class.state_class is None
    assert "Quirks provided an invalid state class: energy" in caplog.text


async def test_cluster_handler_quirks_attribute_reporting(zha_gateway: Gateway) -> None:
    """Test quirks sensor setting up ZCL_INIT_ATTRS and REPORT_CONFIG correctly."""

    # Suppress normal endpoint probing, as this will claim the Opple cluster handler
    # already due to it being in the "CLUSTER_HANDLER_ONLY_CLUSTERS" registry.
    # We want to test the handler also gets claimed via quirks v2 reporting config.
    with patch("zha.application.discovery.discover_entities_for_endpoint"):
        zha_device, cluster = await zigpy_device_aqara_sensor_v2_mock(zha_gateway)

    assert isinstance(zha_device.device, CustomDeviceV2)

    # get cluster handler of OppleCluster
    opple_ch = zha_device.endpoints[1].all_cluster_handlers["1:0xfcc0"]
    assert isinstance(opple_ch, OppleRemoteClusterHandler)

    # make sure the cluster handler was claimed due to reporting config, so ZHA binds it
    assert opple_ch in zha_device.endpoints[1].claimed_cluster_handlers.values()

    # check that BIND is not set to False, as reporting is configured
    assert opple_ch.BIND is True

    # check ZCL_INIT_ATTRS contains sensor attributes that are not in REPORT_CONFIG
    assert opple_ch.ZCL_INIT_ATTRS == {
        "energy": True,
        "energy_delivered": True,
        "energy_invalid_state_class": True,
        "power": True,
    }
    # check that ZCL_INIT_ATTRS is an instance variable and not a class variable now
    assert opple_ch.ZCL_INIT_ATTRS is opple_ch.__dict__[ZCL_INIT_ATTRS]
    assert opple_ch.ZCL_INIT_ATTRS is not OppleRemoteClusterHandler.ZCL_INIT_ATTRS

    # double check we didn't modify the class variable
    assert OppleRemoteClusterHandler.ZCL_INIT_ATTRS == {}

    # check if REPORT_CONFIG is set correctly
    assert (
        (
            AttrReportConfig(
                attr="last_feeding_size",
                config=(0, 60, 1),
            ),
        )
    ) == opple_ch.REPORT_CONFIG

    # this cannot be wrong, as REPORT_CONFIG is an immutable tuple and not a list/dict,
    # but let's check it anyway in case the type changes in the future
    assert opple_ch.REPORT_CONFIG is not OppleRemoteClusterHandler.REPORT_CONFIG
    assert OppleRemoteClusterHandler.REPORT_CONFIG == ()


async def test_cluster_handler_quirks_attribute_reading(zha_gateway: Gateway) -> None:
    """Test quirks sensor setting up ZCL_INIT_ATTRS, claiming cluster handler."""

    registry = DeviceRegistry()
    (
        QuirkBuilder(
            "Fake_Manufacturer_sensor_2", "Fake_Model_sensor_2", registry=registry
        )
        .replaces(OppleCluster)
        .sensor(
            "last_feeding_size",
            OppleCluster.cluster_id,
            translation_key="last_feeding_size",
            fallback_name="Last feeding size",
        )
        .add_to_registry()
    )

    zigpy_device = create_mock_zigpy_device(
        zha_gateway,
        {
            1: {
                SIG_EP_INPUT: [
                    general.Basic.cluster_id,
                    OppleCluster.cluster_id,
                ],
                SIG_EP_OUTPUT: [],
                SIG_EP_TYPE: zigpy.profiles.zha.DeviceType.OCCUPANCY_SENSOR,
                SIG_EP_PROFILE: zigpy.profiles.zha.PROFILE_ID,
            }
        },
        manufacturer="Fake_Manufacturer_sensor_2",
        model="Fake_Model_sensor_2",
    )
    zigpy_device = registry.get_device(zigpy_device)

    # Suppress normal endpoint probing, as this will claim the Opple cluster handler
    # already due to it being in the "CLUSTER_HANDLER_ONLY_CLUSTERS" registry.
    # We want to test the handler also gets claimed via quirks v2 attributes init.
    with patch("zha.application.discovery.discover_entities_for_endpoint"):
        zha_device = await join_zigpy_device(zha_gateway, zigpy_device)

    assert isinstance(zha_device.device, CustomDeviceV2)

    # get cluster handler of OppleCluster
    opple_ch = zha_device.endpoints[1].all_cluster_handlers["1:0xfcc0"]
    assert isinstance(opple_ch, OppleRemoteClusterHandler)

    # make sure the cluster handler was claimed due to attributes to be initialized
    # otherwise, ZHA won't configure the cluster handler, so attributes are not read
    assert opple_ch in zha_device.endpoints[1].claimed_cluster_handlers.values()

    # check that BIND is set to False, as no reporting is configured
    assert opple_ch.BIND is False

    # check ZCL_INIT_ATTRS contains sensor attributes that are not in REPORT_CONFIG
    assert opple_ch.ZCL_INIT_ATTRS == {
        "last_feeding_size": True,
    }
    # check that ZCL_INIT_ATTRS is an instance variable and not a class variable now
    assert opple_ch.ZCL_INIT_ATTRS is opple_ch.__dict__[ZCL_INIT_ATTRS]
    assert opple_ch.ZCL_INIT_ATTRS is not OppleRemoteClusterHandler.ZCL_INIT_ATTRS

    # double check we didn't modify the class variable
    assert OppleRemoteClusterHandler.ZCL_INIT_ATTRS == {}

    # check if REPORT_CONFIG is empty, both instance and class variable
    assert opple_ch.REPORT_CONFIG == ()
    assert OppleRemoteClusterHandler.REPORT_CONFIG == ()


async def test_device_counter_sensors(zha_gateway: Gateway) -> None:
    """Test coordinator counter sensor."""

    coordinator = zha_gateway.coordinator_zha_device
    assert coordinator.is_coordinator
    entity = get_entity(
        coordinator,
        platform=Platform.SENSOR,
        qualifier_func=lambda e: e.info_object.unique_id.endswith(
            "ezsp_counters_counter_1"
        ),
    )

    assert entity.state["state"] == 1

    # simulate counter increment on application
    coordinator.device.application.state.counters["ezsp_counters"][
        "counter_1"
    ].increment()

    await asyncio.sleep(zha_gateway.global_updater.__polling_interval + 2)
    await zha_gateway.async_block_till_done(wait_background_tasks=True)

    assert entity.state["state"] == 2

    # test disabling the entity disables it and removes it from the updater
    assert len(zha_gateway.global_updater._update_listeners) == 3
    assert entity.enabled is True

    entity.disable()

    assert entity.enabled is False
    assert len(zha_gateway.global_updater._update_listeners) == 2

    # test enabling the entity enables it and adds it to the updater
    entity.enable()

    assert entity.enabled is True
    assert len(zha_gateway.global_updater._update_listeners) == 3

    # make sure we don't get multiple listeners for the same entity in the updater
    entity.enable()

    assert len(zha_gateway.global_updater._update_listeners) == 3


async def test_device_unavailable_or_disabled_skips_entity_polling(
    zha_gateway: Gateway,
    caplog: pytest.LogCaptureFixture,
) -> None:
    """Test polling is skipped for unavailable devices."""

    elec_measurement_zigpy_dev = elec_measurement_zigpy_device_mock(zha_gateway)
    elec_measurement_zha_dev = await join_zigpy_device(
        zha_gateway, elec_measurement_zigpy_dev
    )
    assert not elec_measurement_zha_dev.is_coordinator
    assert not elec_measurement_zha_dev.is_active_coordinator
    entity = get_entity(
        elec_measurement_zha_dev,
        platform=Platform.SENSOR,
        exact_entity_type=sensor.RSSISensor,
    )

    assert entity.state["state"] is None

    elec_measurement_zha_dev.device.rssi = 60

    await asyncio.sleep(zha_gateway.global_updater.__polling_interval + 2)
    await zha_gateway.async_block_till_done(wait_background_tasks=True)

    assert entity.state["state"] == 60
    assert entity.enabled is True
    assert len(zha_gateway.global_updater._update_listeners) == 5

    # let's drop the normal update method from the updater
    entity.disable()

    assert entity.enabled is False
    assert len(zha_gateway.global_updater._update_listeners) == 4

    # wrap the update method so we can count how many times it was called
    entity.update = MagicMock(wraps=entity.update)
    await asyncio.sleep(zha_gateway.global_updater.__polling_interval + 2)
    await zha_gateway.async_block_till_done(wait_background_tasks=True)

    assert entity.update.call_count == 0

    # re-enable the entity and ensure it is back in the updater and that update is called
    entity.enable()
    assert len(zha_gateway.global_updater._update_listeners) == 5
    assert entity.enabled is True

    await asyncio.sleep(zha_gateway.global_updater.__polling_interval + 2)
    await zha_gateway.async_block_till_done(wait_background_tasks=True)

    assert entity.update.call_count == 1

    # knock it off the network and ensure the polling is skipped
    assert (
        "00:0d:6f:00:0a:90:69:e7-1-0-rssi: skipping polling for updated state, "
        "available: False, allow polled requests: True" not in caplog.text
    )

    elec_measurement_zha_dev.on_network = False
    await asyncio.sleep(zha_gateway.global_updater.__polling_interval + 2)
    await zha_gateway.async_block_till_done(wait_background_tasks=True)

    assert entity.update.call_count == 2

    assert (
        "00:0d:6f:00:0a:90:69:e7-1-0-rssi: skipping polling for updated state, "
        "available: False, allow polled requests: True" in caplog.text
    )


async def zigpy_device_danfoss_thermostat_mock(
    zha_gateway: Gateway,
) -> tuple[Device, zigpy.device.Device]:
    """Danfoss thermostat device."""

    zigpy_device = create_mock_zigpy_device(
        zha_gateway,
        {
            1: {
                SIG_EP_INPUT: [
                    general.Basic.cluster_id,
                    general.PowerConfiguration.cluster_id,
                    general.Identify.cluster_id,
                    general.Time.cluster_id,
                    general.PollControl.cluster_id,
                    hvac.Thermostat.cluster_id,
                    hvac.UserInterface.cluster_id,
                    homeautomation.Diagnostic.cluster_id,
                ],
                SIG_EP_OUTPUT: [general.Basic.cluster_id, general.Ota.cluster_id],
                SIG_EP_TYPE: zigpy.profiles.zha.DeviceType.THERMOSTAT,
                SIG_EP_PROFILE: zigpy.profiles.zha.PROFILE_ID,
            }
        },
        manufacturer="Danfoss",
        model="eTRV0100",
        quirk=danfoss_thermostat.DanfossThermostat,
    )

    zha_device = await join_zigpy_device(zha_gateway, zigpy_device)
    return zha_device, zigpy_device


async def test_danfoss_thermostat_sw_error(zha_gateway: Gateway) -> None:
    """Test quirks defined thermostat."""

    zha_device, zigpy_device = await zigpy_device_danfoss_thermostat_mock(zha_gateway)

    entity = get_entity(
        zha_device,
        platform=Platform.SENSOR,
        exact_entity_type=DanfossSoftwareErrorCode,
        qualifier="sw_error_code",
    )
    assert entity is not None

    cluster = zigpy_device.endpoints[1].diagnostic

    await send_attributes_report(
        zha_gateway,
        cluster,
        {
            danfoss_thermostat.DanfossDiagnosticCluster.AttributeDefs.sw_error_code.id: 0x0001
        },
    )

    assert entity.state["state"] == "something"
    assert entity.extra_state_attribute_names
    assert "Top_pcb_sensor_error" in entity.extra_state_attribute_names
    assert entity.state["Top_pcb_sensor_error"]


async def test_quirks_sensor_attr_converter(zha_gateway: Gateway) -> None:
    """Test ZHA quirks v2 sensor with attribute_converter."""

    registry = DeviceRegistry()
    zigpy_dev = create_mock_zigpy_device(
        zha_gateway,
        {
            1: {
                SIG_EP_INPUT: [
                    general.Basic.cluster_id,
                    general.AnalogInput.cluster_id,
                ],
                SIG_EP_OUTPUT: [],
                SIG_EP_TYPE: zha.DeviceType.SIMPLE_SENSOR,
                SIG_EP_PROFILE: zigpy.profiles.zha.PROFILE_ID,
            }
        },
        manufacturer="manufacturer",
        model="model",
    )

    (
        QuirkBuilder(zigpy_dev.manufacturer, zigpy_dev.model, registry=registry)
        .sensor(
            AnalogInput.AttributeDefs.present_value.name,
            AnalogInput.cluster_id,
            translation_key="quirks_sensor",
            fallback_name="Quirks sensor",
            attribute_converter=lambda x: x + 100,
        )
        .add_to_registry()
    )

    zigpy_device_ = registry.get_device(zigpy_dev)

    assert isinstance(zigpy_device_, CustomDeviceV2)
    cluster = zigpy_device_.endpoints[1].analog_input

    zha_device = await join_zigpy_device(zha_gateway, zigpy_device_)
    entity = get_entity(zha_device, platform=Platform.SENSOR, qualifier="present_value")

    # send updated value, check if the value is converted
    await send_attributes_report(zha_gateway, cluster, {"present_value": 100})
    assert entity.state["state"] == 200.0

    await send_attributes_report(zha_gateway, cluster, {"present_value": 0})
    assert entity.state["state"] == 100.0


async def test_ignore_non_value(zha_gateway: Gateway) -> None:
    """Test sensor updates ignoring ZCL datatype non-values."""

    zigpy_dev = await zigpy_device_from_json(
        zha_gateway.application_controller,
        "tests/data/devices/third-reality-inc-3rsm0147z.json",
    )

    zha_device = await join_zigpy_device(zha_gateway, zigpy_dev)
    cluster = zha_device.device.endpoints[1].temperature
    entity = get_entity(zha_device, platform=Platform.SENSOR, entity_type=Temperature)

    assert entity.state["state"] == 22.3

    # Normal attribute report
    await send_attributes_report(
        zha_gateway,
        cluster,
        {measurement.TemperatureMeasurement.AttributeDefs.measured_value.id: 3000},
    )
    assert entity.state["state"] == 30.0

    # Invalid attribute value, ignored
    await send_attributes_report(
        zha_gateway,
        cluster,
        {measurement.TemperatureMeasurement.AttributeDefs.measured_value.id: -0x8000},
    )
    assert entity.state["state"] is None


@pytest.mark.parametrize(
    ("resolution", "precision"),
    [
        (0.0009999999310821295, 3),
        (0.001, 3),
        (0.001000001, 3),
        (0.01, 2),
        (0.05, 2),
        (0.1, 1),
        (1.0, 0),
        (2.0, 0),
        (0.9, 1),
        (0.7, 1),
        (0.2, 1),
        (0.25, 2),
        (0.33, 2),
        (0.44, 2),
        (0.55, 2),
        (0.66, 2),
        (0.77, 2),
        (0.88, 2),
        (1 / 3, 7),
        (0.125, 3),
        (0.5, 1),
        (1.5, 1),
        (10.5, 1),
        (0.000000000000001, 15),
        (0.0000000001, 10),
        (0.000000001, 9),
        (0.00000001, 8),
        (0.0000001, 7),
        (0.000001, 6),
    ],
)
def test_sensor_precision(resolution: float, precision: int) -> None:
    """Test converting ZCL `resolution` into precision."""

    # Force the values to be float32, not float64
    resolution, _ = t.Single.deserialize(t.Single(resolution).serialize())
    assert resolution_to_decimal_precision(resolution) == precision


async def test_enum_sensor(zha_gateway: Gateway) -> None:
    """Test enum sensor creation and handling of missing values."""
    registry = DeviceRegistry()
    zigpy_dev = await zigpy_device_from_json(
        zha_gateway.application_controller,
        "tests/data/devices/third-reality-inc-3rsm0147z.json",
    )

    zigpy_dev.endpoints[1].power.update_attribute(
        PowerConfiguration.AttributeDefs.battery_size.id,
        PowerConfiguration.BatterySize.AAA,
    )

    (
        QuirkBuilder(zigpy_dev.manufacturer, zigpy_dev.model, registry=registry)
        .enum(
            entity_platform=EntityPlatform.SENSOR,
            entity_type=EntityType.DIAGNOSTIC,
            attribute_name=PowerConfiguration.AttributeDefs.battery_size.name,
            enum_class=PowerConfiguration.BatterySize,
            cluster_id=PowerConfiguration.cluster_id,
            endpoint_id=1,
            translation_key="battery_size",
            fallback_name="Battery size",
        )
        .add_to_registry()
    )

    zigpy_dev = registry.get_device(zigpy_dev)

    zha_device = await join_zigpy_device(zha_gateway, zigpy_dev)
    entity = get_entity(zha_device, platform=Platform.SENSOR, qualifier="battery_size")

    assert entity.state["state"] == "AAA"

    zigpy_dev.endpoints[1].power.update_attribute(
        PowerConfiguration.AttributeDefs.battery_size.id,
        0xAB,  # unknown
    )

    assert entity.state["state"] == "undefined_0xab"  # TODO: should this be `None`?



"""Device discovery functions for Zigbee Home Automation."""

from __future__ import annotations

from collections import Counter, defaultdict
from collections.abc import Callable, Iterator
from dataclasses import astuple
import functools
import itertools
import logging
from typing import TYPE_CHECKING

from zigpy.quirks.v2 import (
    BinarySensorMetadata,
    CustomDeviceV2,
    NumberMetadata,
    SwitchMetadata,
    WriteAttributeButtonMetadata,
    ZCLCommandButtonMetadata,
    ZCLEnumMetadata,
    ZCLSensorMetadata,
)
from zigpy.state import State
from zigpy.zcl import ClusterType

from zha.application import Platform, const as zha_const
from zha.application.platforms import (  # noqa: F401 pylint: disable=unused-import
    ENTITY_REGISTRY,
    GROUP_ENTITY_REGISTRY,
    BaseEntity,
    ClusterHandlerMatch,
    PlatformEntity,
    PlatformFeatureGroup,
    alarm_control_panel,
    binary_sensor,
    button,
    climate,
    cover,
    device_tracker,
    fan,
    light,
    lock,
    number,
    select,
    sensor,
    siren,
    switch,
    update,
)

# importing cluster handlers updates registries
from zha.zigbee.cluster_handlers import (  # noqa: F401 pylint: disable=unused-import
    AttrReportConfig,
    ClientClusterHandler,
    ClusterHandler,
    closures,
    general,
    homeautomation,
    hvac,
    lighting,
    lightlink,
    manufacturerspecific,
    measurement,
    protocol,
    security,
    smartenergy,
)
from zha.zigbee.group import Group

if TYPE_CHECKING:
    from zha.application.platforms import GroupEntity
    from zha.zigbee.device import Device
    from zha.zigbee.endpoint import Endpoint

_LOGGER = logging.getLogger(__name__)

PLATFORMS = (
    Platform.ALARM_CONTROL_PANEL,
    Platform.BINARY_SENSOR,
    Platform.BUTTON,
    Platform.CLIMATE,
    Platform.COVER,
    Platform.DEVICE_TRACKER,
    Platform.FAN,
    Platform.LIGHT,
    Platform.LOCK,
    Platform.NUMBER,
    Platform.SELECT,
    Platform.SENSOR,
    Platform.SIREN,
    Platform.SWITCH,
    Platform.UPDATE,
)

GROUP_PLATFORMS = (
    Platform.FAN,
    Platform.LIGHT,
    Platform.SWITCH,
)

QUIRKS_ENTITY_META_TO_ENTITY_CLASS = {
    (Platform.BUTTON, WriteAttributeButtonMetadata): button.WriteAttributeButton,
    (Platform.BUTTON, ZCLCommandButtonMetadata): button.Button,
    (Platform.BINARY_SENSOR, BinarySensorMetadata): binary_sensor.BinarySensor,
    (Platform.SENSOR, ZCLEnumMetadata): sensor.EnumSensor,
    (Platform.SENSOR, ZCLSensorMetadata): sensor.Sensor,
    (Platform.SELECT, ZCLEnumMetadata): select.ZCLEnumSelectEntity,
    (Platform.NUMBER, NumberMetadata): number.NumberConfigurationEntity,
    (Platform.SWITCH, SwitchMetadata): switch.ConfigurableAttributeSwitch,
}


def ignore_exceptions_during_iteration[**P, T](
    func: Callable[P, Iterator[T]],
) -> Callable[P, Iterator[T]]:
    """Ignore exceptions during iteration for wrapped function."""

    @functools.wraps(func)
    def inner(*args: P.args, **kwargs: P.kwargs) -> Iterator[T]:
        iterator = func(*args, **kwargs)
        while True:
            try:
                yield next(iterator)
            except StopIteration:
                break
            except Exception:  # pylint: disable=broad-except
                _LOGGER.exception("Failed to create entity during discovery")

    return inner


@ignore_exceptions_during_iteration
def discover_device_entities(device: Device) -> Iterator[BaseEntity]:
    """Discover entities for a ZHA device."""
    _LOGGER.debug(
        "Discovering entities for device: %s-%s",
        str(device.ieee),
        device.name,
    )

    assert not device.is_active_coordinator

    for ep_id, endpoint in device.endpoints.items():
        if ep_id == 0:
            continue

        _LOGGER.debug(
            "Discovering entities for endpoint: %s-%s",
            str(endpoint.device.ieee),
            endpoint.id,
        )

        yield from discover_entities_for_endpoint(endpoint)

    yield from discover_quirks_v2_entities(device)


@ignore_exceptions_during_iteration
def discover_coordinator_device_entities(
    device: Device,
) -> Iterator[sensor.DeviceCounterSensor]:
    """Discover entities for the coordinator device."""
    _LOGGER.debug(
        "Discovering entities for coordinator device: %s-%s",
        str(device.ieee),
        device.name,
    )
    state: State = device.gateway.application_controller.state

    for counter_groups in (
        "counters",
        "broadcast_counters",
        "device_counters",
        "group_counters",
    ):
        for counter_group, counters in getattr(state, counter_groups).items():
            for counter in counters:
                yield sensor.DeviceCounterSensor(
                    zha_device=device,
                    counter_groups=counter_groups,
                    counter_group=counter_group,
                    counter=counter,
                )

                _LOGGER.debug(
                    "'%s' platform -> '%s' using %s",
                    Platform.SENSOR,
                    sensor.DeviceCounterSensor.__name__,
                    f"counter groups[{counter_groups}] counter group[{counter_group}] counter[{counter}]",
                )


@ignore_exceptions_during_iteration
def discover_group_entities(group: Group) -> Iterator[GroupEntity]:
    """Process a group and create any entities that are needed."""
    # only create a group entity if there are 2 or more members in a group
    if len(group.members) < 2:
        _LOGGER.debug(
            "Group: %s:0x%04x has less than 2 members - skipping entity discovery",
            group.name,
            group.group_id,
        )
        group.group_entities.clear()
        return

    # We only create groups with two or more devices
    platform_counts: Counter[Platform] = Counter()

    for member in group.members:
        if member.device.is_coordinator:
            continue

        for entity in member.associated_entities:
            platform_counts[entity.PLATFORM] += 1

    for platform, count in platform_counts.items():
        if count < 2:
            continue

        for group_entity_class in GROUP_ENTITY_REGISTRY:
            if platform != group_entity_class.PLATFORM:
                continue
            _LOGGER.info(
                "Creating group entity %s for group %s",
                group_entity_class,
                group.name,
            )
            yield group_entity_class(group)


def discover_quirks_v2_entities(device: Device) -> Iterator[PlatformEntity]:
    """Discover entities for a ZHA device exposed by quirks v2."""
    _LOGGER.debug(
        "Attempting to discover quirks v2 entities for device: %s-%s",
        str(device.ieee),
        device.name,
    )

    if not isinstance(device.device, CustomDeviceV2):
        _LOGGER.debug(
            "Device: %s-%s is not a quirks v2 device - skipping "
            "discover_quirks_v2_entities",
            str(device.ieee),
            device.name,
        )
        return

    zigpy_device: CustomDeviceV2 = device.device

    if not zigpy_device.exposes_metadata:
        _LOGGER.debug(
            "Device: %s-%s does not expose any quirks v2 entities",
            str(device.ieee),
            device.name,
        )
        return

    for (
        cluster_details,
        entity_metadata_list,
    ) in zigpy_device.exposes_metadata.items():
        endpoint_id, cluster_id, cluster_type = cluster_details

        if endpoint_id not in device.endpoints:
            _LOGGER.warning(
                "Device: %s-%s does not have an endpoint with id: %s - unable to "
                "create entity with cluster details: %s",
                str(device.ieee),
                device.name,
                endpoint_id,
                cluster_details,
            )
            continue

        endpoint: Endpoint = device.endpoints[endpoint_id]
        cluster = (
            endpoint.zigpy_endpoint.in_clusters.get(cluster_id)
            if cluster_type is ClusterType.Server
            else endpoint.zigpy_endpoint.out_clusters.get(cluster_id)
        )

        if cluster is None:
            _LOGGER.warning(
                "Device: %s-%s does not have a cluster with id: %s - "
                "unable to create entity with cluster details: %s",
                str(device.ieee),
                device.name,
                cluster_id,
                cluster_details,
            )
            continue

        if cluster_type is ClusterType.Server:
            cluster_handler = endpoint.all_cluster_handlers.get(
                f"{endpoint.id}:0x{cluster.cluster_id:04x}"
            )
        else:
            cluster_handler = endpoint.client_cluster_handlers.get(
                f"{endpoint.id}:0x{cluster.cluster_id:04x}_client"
            )

        assert cluster_handler

        # flags to determine if we need to claim/bind the cluster handler
        attribute_initialization_found: bool = False
        reporting_found: bool = False

        for entity_metadata in entity_metadata_list:
            platform = Platform(entity_metadata.entity_platform.value)
            metadata_type = type(entity_metadata)
            entity_class = QUIRKS_ENTITY_META_TO_ENTITY_CLASS.get(
                (platform, metadata_type)
            )

            if entity_class is None:
                _LOGGER.warning(
                    "Device: %s-%s has an entity with details: %s that does not"
                    " have an entity class mapping - unable to create entity",
                    str(device.ieee),
                    device.name,
                    {
                        zha_const.CLUSTER_DETAILS: cluster_details,
                        zha_const.ENTITY_METADATA: entity_metadata,
                    },
                )
                continue

            # process the entity metadata for ZCL_INIT_ATTRS and REPORT_CONFIG
            if attr_name := getattr(entity_metadata, "attribute_name", None):
                # TODO: ignore "attribute write buttons"? currently, we claim ch
                # if the entity has a reporting config, add it to the cluster handler
                if rep_conf := getattr(entity_metadata, "reporting_config", None):
                    # if attr is already in REPORT_CONFIG, remove it first
                    cluster_handler.REPORT_CONFIG = tuple(
                        filter(
                            lambda cfg: cfg["attr"] != attr_name,
                            cluster_handler.REPORT_CONFIG,
                        )
                    )
                    # tuples are immutable and we re-set the REPORT_CONFIG here,
                    # so no need to check for an instance variable
                    cluster_handler.REPORT_CONFIG += (
                        AttrReportConfig(attr=attr_name, config=astuple(rep_conf)),
                    )
                    # mark cluster handler for claiming and binding later
                    reporting_found = True

                # not in REPORT_CONFIG, add to ZCL_INIT_ATTRS if it not already in
                elif attr_name not in cluster_handler.ZCL_INIT_ATTRS:
                    # copy existing ZCL_INIT_ATTRS into instance variable once,
                    # so we don't modify other instances of the same cluster handler
                    if zha_const.ZCL_INIT_ATTRS not in cluster_handler.__dict__:
                        cluster_handler.ZCL_INIT_ATTRS = (
                            cluster_handler.ZCL_INIT_ATTRS.copy()
                        )
                    # add the attribute to the guaranteed instance variable
                    cluster_handler.ZCL_INIT_ATTRS[attr_name] = (
                        entity_metadata.attribute_initialized_from_cache
                    )
                    # mark cluster handler for claiming later, but not binding
                    attribute_initialization_found = True

            yield entity_class(
                cluster_handlers=[cluster_handler],
                endpoint=endpoint,
                device=device,
                entity_metadata=entity_metadata,
                legacy_discovery_unique_id=f"{device.ieee}-{endpoint.id}",
            )

            _LOGGER.debug(
                "'%s' platform -> '%s' using %s",
                platform,
                entity_class.__name__,
                [cluster_handler.name],
            )

        # if the cluster handler is unclaimed, claim it and set BIND accordingly,
        # so ZHA configures the cluster handler: reporting + reads attributes
        if (attribute_initialization_found or reporting_found) and (
            cluster_handler not in endpoint.claimed_cluster_handlers.values()
        ):
            endpoint.claim_cluster_handlers([cluster_handler])
            # BIND is True by default, so only set to False if no reporting found.
            # We can safely do this, since quirks v2 entities are initialized last,
            # so if the cluster handler wasn't claimed by endpoint probing so far,
            # only v2 entities need it.
            if not reporting_found:
                cluster_handler.BIND = False


def discover_entities_for_endpoint(endpoint: Endpoint) -> Iterator[PlatformEntity]:
    """Discover entities for an endpoint using the new registry-based discovery."""
    device = endpoint.device

    # TODO: deprecate device platform overrides. The only use case is to swap between
    # `light` and `switch` for devices whose device type is incorrect.
    platform_override: Platform | None = None

    if (
        device_override := device.gateway.config.config.device_overrides.get(
            f"{device.ieee}-{endpoint.id}"
        )
    ) is not None:
        platform_override = device_override.type

    matches_by_feature_and_priority: defaultdict[
        PlatformFeatureGroup | None,
        defaultdict[
            int,  # Weight
            list[tuple[ClusterHandlerMatch, type[PlatformEntity]]],
        ],
    ] = defaultdict(lambda: defaultdict(list))

    for cluster in itertools.chain(
        endpoint.zigpy_endpoint.in_clusters.values(),
        endpoint.zigpy_endpoint.out_clusters.values(),
    ):
        # To speed up lookups, we key ENTITY_REGISTRY by cluster ID. First, we find all
        # compatible entities and their matching criteria.
        for entity_class in ENTITY_REGISTRY.get(cluster.cluster_id, []):
            match = entity_class._cluster_handler_match
            if match is None:
                continue

            if not match.cluster_handlers.issubset(
                endpoint.cluster_handlers_by_name.keys()
            ):
                continue

            if not match.client_cluster_handlers.issubset(
                endpoint.client_cluster_handlers_by_name.keys()
            ):
                continue

            if (
                match.exposed_features is not None
                and not match.exposed_features & device.exposes_features
            ):
                continue

            if (
                match.manufacturers is not None
                and device.manufacturer not in match.manufacturers
            ):
                continue

            if match.models is not None and device.model not in match.models:
                continue

            profile_device_type = (
                endpoint.zigpy_endpoint.profile_id,
                endpoint.zigpy_endpoint.device_type,
            )
            if (
                match.profile_device_types is not None
                and profile_device_type not in match.profile_device_types
            ):
                continue

            if (
                match.not_profile_device_types is not None
                and profile_device_type in match.not_profile_device_types
            ):
                continue

            if match.feature_priority is not None:
                feature, priority = match.feature_priority
            else:
                feature = None
                priority = 0

            # Finally, account for `platform_override` by boosting the priority of
            # matching platforms
            if (
                platform_override is not None
                and platform_override == entity_class.PLATFORM
            ):
                priority += 1000

            matches_by_feature_and_priority[feature][priority].append(
                (match, entity_class)
            )

    # Then, we process the matches and discard entities with lower weights (when
    # feature groups are used)
    for feature, matches_by_priority in matches_by_feature_and_priority.items():
        highest_priority = max(matches_by_priority.keys())

        if _LOGGER.getEffectiveLevel() <= logging.DEBUG:
            ignored_matches = [
                (priority, matches)
                for priority, matches in matches_by_priority.items()
                if priority < highest_priority
            ]

            if ignored_matches:
                _LOGGER.debug(
                    "Ignored matches for feature '%s': %s",
                    feature,
                    ignored_matches,
                )

        for match, entity_class in matches_by_priority[highest_priority]:
            server_handlers = set(match.cluster_handlers)

            for optional in match.optional_cluster_handlers:
                if optional in endpoint.cluster_handlers_by_name:
                    server_handlers.add(optional)

            client_handlers = set(match.client_cluster_handlers)

            server_cluster_handlers = [
                endpoint.cluster_handlers_by_name[name] for name in server_handlers
            ]
            client_cluster_handlers = [
                endpoint.client_cluster_handlers_by_name[name]
                for name in client_handlers
            ]

            # Claim on endpoint
            endpoint.claim_cluster_handlers(server_cluster_handlers)
            endpoint.claim_cluster_handlers(client_cluster_handlers)

            _LOGGER.debug(
                "'%s' platform -> '%s' using %s + %s",
                entity_class.PLATFORM,
                entity_class.__name__,
                [ch.name for ch in server_cluster_handlers],
                [ch.name for ch in client_cluster_handlers],
            )

            # XXX: Combining server and client cluster handlers should not be done
            cluster_handlers: list[ClusterHandler | ClientClusterHandler] = (
                server_cluster_handlers + client_cluster_handlers  # type: ignore[operator]
            )

            yield entity_class(
                cluster_handlers=cluster_handlers,
                endpoint=endpoint,
                device=device,
            )



"""Constants for the lock platform."""

from typing import Final

STATE_LOCKED: Final[str] = "locked"
STATE_UNLOCKED: Final[str] = "unlocked"
STATE_LOCKING: Final[str] = "locking"
STATE_UNLOCKING: Final[str] = "unlocking"
STATE_JAMMED: Final[str] = "jammed"
# The first state is Zigbee 'Not fully locked'
STATE_LIST: Final[list[str]] = [STATE_UNLOCKED, STATE_LOCKED, STATE_UNLOCKED]
VALUE_TO_STATE: Final = dict(enumerate(STATE_LIST))



"""Lights on Zigbee Home Automation networks."""

# pylint: disable=too-many-lines

from __future__ import annotations

from abc import ABC
import asyncio
from collections import Counter
import contextlib
import dataclasses
from dataclasses import dataclass
import functools
import itertools
import logging
from typing import TYPE_CHECKING, Any, cast

from zigpy.zcl.clusters.general import Identify, LevelControl, OnOff
from zigpy.zcl.clusters.lighting import Color
from zigpy.zcl.foundation import Status

from zha.application import Platform
from zha.application.platforms import (
    DEFAULT_UPDATE_GROUP_FROM_CHILD_DELAY,
    BaseEntity,
    BaseEntityInfo,
    ClusterHandlerMatch,
    GroupEntity,
    PlatformEntity,
    PlatformFeatureGroup,
    register_entity,
    register_group_entity,
)
from zha.application.platforms.helpers import (
    find_state_attributes,
    mean_tuple,
    reduce_attribute,
)
from zha.application.platforms.light.const import (
    ASSUME_UPDATE_GROUP_FROM_CHILD_DELAY,
    ATTR_BRIGHTNESS,
    ATTR_COLOR_MODE,
    ATTR_COLOR_TEMP,
    ATTR_EFFECT,
    ATTR_EFFECT_LIST,
    ATTR_MAX_MIREDS,
    ATTR_MIN_MIREDS,
    ATTR_SUPPORTED_COLOR_MODES,
    ATTR_SUPPORTED_FEATURES,
    ATTR_XY_COLOR,
    DEFAULT_EXTRA_TRANSITION_DELAY_LONG,
    DEFAULT_EXTRA_TRANSITION_DELAY_SHORT,
    DEFAULT_LONG_TRANSITION_TIME,
    DEFAULT_MIN_BRIGHTNESS,
    DEFAULT_MIN_TRANSITION_MANUFACTURERS,
    DEFAULT_ON_OFF_TRANSITION,
    EFFECT_COLORLOOP,
    EFFECT_OFF,
    FLASH_EFFECTS,
    LIGHT_PROFILE_DEVICE_TYPES,
    ColorMode,
    FlashMode,
    LightEntityFeature,
)
from zha.application.platforms.light.helpers import (
    filter_supported_color_modes,
    is_brightness_supported,
)
from zha.debounce import Debouncer
from zha.decorators import periodic
from zha.zigbee.cluster_handlers import ClusterAttributeUpdatedEvent
from zha.zigbee.cluster_handlers.const import (
    CLUSTER_HANDLER_ATTRIBUTE_UPDATED,
    CLUSTER_HANDLER_COLOR,
    CLUSTER_HANDLER_LEVEL,
    CLUSTER_HANDLER_LEVEL_CHANGED,
    CLUSTER_HANDLER_ON_OFF,
)
from zha.zigbee.cluster_handlers.general import (
    IdentifyClusterHandler,
    LevelChangeEvent,
    LevelControlClusterHandler,
    OnOffClusterHandler,
)
from zha.zigbee.cluster_handlers.lighting import ColorClusterHandler

if TYPE_CHECKING:
    from zha.application.gateway import Gateway
    from zha.zigbee.cluster_handlers import ClusterHandler
    from zha.zigbee.device import Device
    from zha.zigbee.endpoint import Endpoint
    from zha.zigbee.group import Group

_LOGGER = logging.getLogger(__name__)


@dataclass(frozen=True, kw_only=True)
class LightEntityInfo(BaseEntityInfo):
    """Light entity info."""

    effect_list: list[str] | None = dataclasses.field(default=None)
    supported_features: LightEntityFeature
    min_mireds: int
    max_mireds: int


class BaseLight(BaseEntity, ABC):
    """Operations common to all light entities."""

    PLATFORM = Platform.LIGHT
    _attr_extra_state_attribute_names: set[str] = {
        "off_with_transition",
        "off_brightness",
    }
    _attr_primary_weight = 10

    def __init__(self, *args, **kwargs):
        """Initialize the light."""
        super().__init__(*args, **kwargs)
        self._min_mireds: int | None = 153
        self._max_mireds: int | None = 500
        self._xy_color: tuple[float, float] | None = None
        self._color_mode = ColorMode.UNKNOWN  # Set by subclasses
        self._color_temp: int | None = None
        self._supported_features: LightEntityFeature = LightEntityFeature(0)
        self._state: bool | None
        self._brightness: int | None = None
        self._off_with_transition: bool = False
        self._off_brightness: int | None = None
        self._effect_list: list[str] | None = None
        self._effect: str = EFFECT_OFF
        self._supported_color_modes: set[ColorMode] = set()

    @property
    def state(self) -> dict[str, Any]:
        """Return the state of the light."""
        response = super().state
        response["on"] = self.is_on
        response["brightness"] = self.brightness
        response["xy_color"] = self.xy_color
        response["color_temp"] = self.color_temp
        response["effect_list"] = self.effect_list
        response["effect"] = self.effect
        response["supported_features"] = self.supported_features
        response["color_mode"] = self.color_mode
        response["supported_color_modes"] = self.supported_color_modes
        response["off_with_transition"] = self._off_with_transition
        response["off_brightness"] = self._off_brightness
        return response

    @functools.cached_property
    def info_object(self) -> LightEntityInfo:
        """Return a representation of the select."""
        return LightEntityInfo(
            **super().info_object.__dict__,
            effect_list=self.effect_list,
            supported_features=self.supported_features,
            min_mireds=self.min_mireds,
            max_mireds=self.max_mireds,
        )

    @property
    def xy_color(self) -> tuple[float, float] | None:
        """Return the xy color value [float, float]."""
        return self._xy_color

    @property
    def color_temp(self) -> int | None:
        """Return the CT color value in mireds."""
        return self._color_temp

    @property
    def color_mode(self) -> ColorMode | None:
        """Return the color mode."""
        return self._color_mode

    @property
    def effect_list(self) -> list[str] | None:
        """Return the list of supported effects."""
        return self._effect_list

    @property
    def effect(self) -> str:
        """Return the current effect."""
        return self._effect

    @property
    def supported_features(self) -> LightEntityFeature:
        """Flag supported features."""
        return self._supported_features

    @property
    def supported_color_modes(self) -> set[ColorMode]:
        """Flag supported color modes."""
        return self._supported_color_modes

    @property
    def is_on(self) -> bool:
        """Return true if entity is on."""
        if self._state is None:
            return False
        return self._state

    @property
    def brightness(self) -> int | None:
        """Return the brightness of this light."""
        return self._brightness

    @property
    def min_mireds(self) -> int | None:
        """Return the coldest color_temp that this light supports."""
        return self._min_mireds

    @property
    def max_mireds(self) -> int | None:
        """Return the warmest color_temp that this light supports."""
        return self._max_mireds

    async def async_turn_on(
        self,
        *,
        transition: float | None = None,
        brightness: int | None = None,
        effect: str | None = None,
        flash: FlashMode | None = None,
        color_temp: int | None = None,
        xy_color: tuple[int, int] | None = None,
    ) -> None:
        """Turn the entity on."""
        raise NotImplementedError

    async def async_turn_off(self, *, transition: float | None = None) -> None:
        """Turn the entity off."""
        raise NotImplementedError

    def restore_external_state_attributes(
        self,
        *,
        state: bool | None,
        off_with_transition: bool | None,
        off_brightness: int | None,
        brightness: int | None,
        color_temp: int | None,
        xy_color: tuple[float, float] | None,
        color_mode: ColorMode | None,
        effect: str | None,
    ) -> None:
        """Restore extra state attributes that are stored outside of the ZCL cache."""
        if state is not None:
            self._state = state
        if off_with_transition is not None:
            self._off_with_transition = off_with_transition
        if off_brightness is not None:
            self._off_brightness = off_brightness
        if brightness is not None:
            self._brightness = brightness
        if color_temp is not None:
            self._color_temp = color_temp
        if xy_color is not None:
            self._xy_color = xy_color
        if color_mode is not None:
            self._color_mode = color_mode
        if effect is not None:
            self._effect = effect


class BaseClusterHandlerLight(BaseLight):
    """Operations common to all light entities."""

    _FORCE_ON = False
    _DEFAULT_MIN_TRANSITION_TIME: float = 0

    def __init__(self, *args, **kwargs):
        """Initialize the light."""
        super().__init__(*args, **kwargs)
        self._zha_config_transition: float = self._DEFAULT_MIN_TRANSITION_TIME
        self._zha_config_enhanced_light_transition: bool = False
        self._zha_config_enable_light_transitioning_flag: bool = True
        self._on_off_cluster_handler: OnOffClusterHandler | None = None
        self._level_cluster_handler: LevelControlClusterHandler | None = None
        self._color_cluster_handler: ColorClusterHandler | None = None
        self._identify_cluster_handler: IdentifyClusterHandler | None = None
        self._transitioning_individual: bool = False
        self._transitioning_group: bool = False
        self._transition_listener: asyncio.TimerHandle | None = None

        self._internal_supported_color_modes: set[ColorMode] = set()

    @property
    def _gateway(self) -> Gateway:
        """Return the gateway."""
        raise NotImplementedError

    @property
    def state(self) -> dict[str, Any]:
        """Return the state of the light."""
        response = super().state
        # XXX: for backwards compatibility
        response["supported_color_modes"] = self._internal_supported_color_modes
        return response

    def recompute_capabilities(self) -> None:
        """Recompute supported features and color modes."""
        super().recompute_capabilities()
        light_options = self._gateway.config.config.light_options
        self._zha_config_transition = light_options.default_light_transition
        self._zha_config_enhanced_light_transition = (
            light_options.enable_enhanced_light_transition
        )
        self._zha_config_enable_light_transitioning_flag = (
            light_options.enable_light_transitioning_flag
        )

    async def async_turn_on(
        self,
        *,
        transition: float | None = None,
        brightness: int | None = None,
        effect: str | None = None,
        flash: FlashMode | None = None,
        color_temp: int | None = None,
        xy_color: tuple[int, int] | None = None,
    ) -> None:
        """Turn the entity on."""
        duration = (
            transition if transition is not None else self._zha_config_transition
        ) or (
            # if 0 is passed in some devices still need the minimum default
            self._DEFAULT_MIN_TRANSITION_TIME
        )

        execute_if_off_supported = (
            self._GROUP_SUPPORTS_EXECUTE_IF_OFF
            if isinstance(self, LightGroup)
            else self._color_cluster_handler
            and self._color_cluster_handler.execute_if_off_supported
        )

        # A device theoretically could lie about having brightness support and omit the
        # actual LevelControl cluster needed to control it
        brightness_supported = (
            is_brightness_supported(self._internal_supported_color_modes)
            and self._level_cluster_handler is not None
        )

        set_transition_flag = (
            brightness_supported or color_temp is not None or xy_color is not None
        ) and self._zha_config_enable_light_transitioning_flag
        transition_time = (
            (
                duration + DEFAULT_EXTRA_TRANSITION_DELAY_SHORT
                if (
                    (brightness is not None or transition is not None)
                    and brightness_supported
                    or (self._off_with_transition and self._off_brightness is not None)
                    or color_temp is not None
                    or xy_color is not None
                )
                else DEFAULT_ON_OFF_TRANSITION + DEFAULT_EXTRA_TRANSITION_DELAY_SHORT
            )
            if set_transition_flag
            else 0
        )

        # If we need to pause attribute report parsing, we'll do so here.
        # After successful calls, we later start a timer to unset the flag after
        # transition_time.
        # - On an error on the first move to level call, we unset the flag immediately
        #   if no previous timer is running.
        # - On an error on subsequent calls, we start the transition timer,
        #   as a brightness call might have come through.
        if set_transition_flag:
            self.async_transition_set_flag()

        # If the light is currently off but a turn_on call with a color/temperature is
        # sent, the light needs to be turned on first at a low brightness level where
        # the light is immediately transitioned to the correct color. Afterwards, the
        # transition is only from the low brightness to the new brightness.
        # Otherwise, the transition is from the color the light had before being turned
        # on to the new color. This can look especially bad with transitions longer than
        # a second. We do not want to do this for devices that need to be forced to use
        # the on command because we would end up with 4 commands sent:
        # move to level, on, color, move to level... We also will not set this
        # if the bulb is already in the desired color mode with the desired color
        # or color temperature.
        new_color_provided_while_off = (
            self._zha_config_enhanced_light_transition
            and not self._FORCE_ON
            and not self._state
            and (
                (
                    color_temp is not None
                    and (
                        self._color_temp != color_temp
                        or self._color_mode != ColorMode.COLOR_TEMP
                    )
                )
                or (
                    xy_color is not None
                    and (self._xy_color != xy_color or self._color_mode != ColorMode.XY)
                )
            )
            and brightness_supported
            and not execute_if_off_supported
        )

        if (
            brightness is None
            and (self._off_with_transition or new_color_provided_while_off)
            and self._off_brightness is not None
        ):
            brightness = self._off_brightness

        if brightness is not None:
            level = min(254, brightness)
        else:
            level = self._brightness or 254

        t_log = {}

        if new_color_provided_while_off:
            assert self._level_cluster_handler is not None

            # If the light is currently off, we first need to turn it on at a low
            # brightness level with no transition.
            # After that, we set it to the desired color/temperature with no transition.
            result = await self._level_cluster_handler.move_to_level_with_on_off(
                level=DEFAULT_MIN_BRIGHTNESS,
                transition_time=int(10 * self._DEFAULT_MIN_TRANSITION_TIME),
            )
            t_log["move_to_level_with_on_off"] = result
            if result[1] is not Status.SUCCESS:
                # First 'move to level' call failed, so if the transitioning delay
                # isn't running from a previous call,
                # the flag can be unset immediately
                if set_transition_flag and not self._transition_listener:
                    self.async_transition_complete()
                self.debug("turned on: %s", t_log)
                return
            # Currently only setting it to "on", as the correct level state will
            # be set at the second move_to_level call
            self._state = True

        if execute_if_off_supported:
            self.debug("handling color commands before turning on/level")
            if not await self.async_handle_color_commands(
                color_temp,
                duration,  # duration is ignored by lights when off
                xy_color,
                new_color_provided_while_off,
                t_log,
            ):
                # Color calls before on/level calls failed,
                # so if the transitioning delay isn't running from a previous call,
                # the flag can be unset immediately
                if set_transition_flag and not self._transition_listener:
                    self.async_transition_complete()
                self.debug("turned on: %s", t_log)
                return

        if (
            (brightness is not None or transition is not None)
            and not new_color_provided_while_off
            and brightness_supported
        ):
            assert self._level_cluster_handler is not None

            result = await self._level_cluster_handler.move_to_level_with_on_off(
                level=level,
                transition_time=int(10 * duration),
            )
            t_log["move_to_level_with_on_off"] = result
            if result[1] is not Status.SUCCESS:
                # First 'move to level' call failed, so if the transitioning delay
                # isn't running from a previous call, the flag can be unset immediately
                if set_transition_flag and not self._transition_listener:
                    self.async_transition_complete()
                self.debug("turned on: %s", t_log)
                return
            self._state = bool(level)
            if level:
                self._brightness = level

        if (
            (brightness is None and transition is None)
            and not new_color_provided_while_off
            or (self._FORCE_ON and brightness != 0)
        ):
            assert self._on_off_cluster_handler is not None

            # since FORCE_ON lights don't turn on with move_to_level_with_on_off,
            # we should call the on command on the on_off cluster
            # if brightness is not 0.
            result = await self._on_off_cluster_handler.on()
            t_log["on_off"] = result
            if result[1] is not Status.SUCCESS:
                # 'On' call failed, but as brightness may still transition
                # (for FORCE_ON lights), we start the timer to unset the flag after
                # the transition_time if necessary.
                self.async_transition_start_timer(transition_time)
                self.debug("turned on: %s", t_log)
                return
            self._state = True

        if not execute_if_off_supported:
            self.debug("handling color commands after turning on/level")
            if not await self.async_handle_color_commands(
                color_temp,
                duration,
                xy_color,
                new_color_provided_while_off,
                t_log,
            ):
                # Color calls failed, but as brightness may still transition,
                # we start the timer to unset the flag
                self.async_transition_start_timer(transition_time)
                self.debug("turned on: %s", t_log)
                return

        if new_color_provided_while_off:
            assert self._level_cluster_handler is not None

            # The light has the correct color, so we can now transition
            # it to the correct brightness level.
            result = await self._level_cluster_handler.move_to_level(
                level=level, transition_time=int(10 * duration)
            )
            t_log["move_to_level_if_color"] = result
            if result[1] is not Status.SUCCESS:
                self.debug("turned on: %s", t_log)
                return
            self._state = bool(level)
            if level:
                self._brightness = level

        # Our light is guaranteed to have just started the transitioning process
        # if necessary, so we start the delay for the transition (to stop parsing
        # attribute reports after the completed transition).
        self.async_transition_start_timer(transition_time)

        if self._color_cluster_handler is not None:
            if effect == EFFECT_COLORLOOP:
                result = await self._color_cluster_handler.color_loop_set(
                    update_flags=(
                        Color.ColorLoopUpdateFlags.Action
                        | Color.ColorLoopUpdateFlags.Direction
                        | Color.ColorLoopUpdateFlags.Time
                    ),
                    action=Color.ColorLoopAction.Activate_from_current_hue,
                    direction=Color.ColorLoopDirection.Increment,
                    time=transition if transition else 7,
                    start_hue=0,
                )
                t_log["color_loop_set"] = result
                self._effect = EFFECT_COLORLOOP
            elif self._effect == EFFECT_COLORLOOP and effect != EFFECT_COLORLOOP:
                result = await self._color_cluster_handler.color_loop_set(
                    update_flags=Color.ColorLoopUpdateFlags.Action,
                    action=Color.ColorLoopAction.Deactivate,
                    direction=Color.ColorLoopDirection.Decrement,
                    time=0,
                    start_hue=0,
                )
                t_log["color_loop_set"] = result
                self._effect = EFFECT_OFF

        if flash is not None:
            assert self._identify_cluster_handler is not None
            result = await self._identify_cluster_handler.trigger_effect(
                effect_id=FLASH_EFFECTS[flash],
                effect_variant=Identify.EffectVariant.Default,
            )
            t_log["trigger_effect"] = result

        self._off_with_transition = False
        self._off_brightness = None
        self.debug("turned on: %s", t_log)
        self.maybe_emit_state_changed_event()

    async def async_turn_off(self, *, transition: float | None = None) -> None:
        """Turn the entity off."""
        brightness_supported = (
            is_brightness_supported(self._internal_supported_color_modes)
            and self._level_cluster_handler is not None
        )

        transition_time = (
            transition or self._DEFAULT_MIN_TRANSITION_TIME
            if transition is not None
            else DEFAULT_ON_OFF_TRANSITION
        ) + DEFAULT_EXTRA_TRANSITION_DELAY_SHORT

        # Start pausing attribute report parsing
        if self._zha_config_enable_light_transitioning_flag:
            self.async_transition_set_flag()

        # is not none looks odd here, but it will override built in bulb
        # transition times if we pass 0 in here
        if transition is not None and brightness_supported:
            assert self._level_cluster_handler is not None

            result = await self._level_cluster_handler.move_to_level_with_on_off(
                level=0,
                transition_time=int(
                    10 * (transition or self._DEFAULT_MIN_TRANSITION_TIME)
                ),
            )
        else:
            assert self._on_off_cluster_handler is not None
            result = await self._on_off_cluster_handler.off()

        # Pause parsing attribute reports until transition is complete
        if self._zha_config_enable_light_transitioning_flag:
            self.async_transition_start_timer(transition_time)
        self.debug("turned off: %s", result)
        if result[1] is not Status.SUCCESS:
            return
        self._state = False

        if brightness_supported and not self._off_with_transition:
            # store current brightness so that the next turn_on uses it:
            # when using "enhanced turn on"
            self._off_brightness = self._brightness
            if transition is not None:
                # save for when calling turn_on without a brightness:
                # current_level is set to 1 after transitioning to level 0,
                # needed for correct state with light groups
                self._brightness = 1
                self._off_with_transition = transition is not None

        self.maybe_emit_state_changed_event()

    async def async_handle_color_commands(
        self,
        color_temp,
        duration,
        xy_color,
        new_color_provided_while_off,
        t_log,
    ):
        """Process ZCL color commands."""
        transition_time = (
            self._DEFAULT_MIN_TRANSITION_TIME
            if new_color_provided_while_off
            else duration
        )

        if color_temp is not None:
            assert self._color_cluster_handler is not None

            result = await self._color_cluster_handler.move_to_color_temp(
                color_temp_mireds=color_temp,
                transition_time=int(10 * transition_time),
            )
            t_log["move_to_color_temp"] = result
            if result[1] is not Status.SUCCESS:
                return False
            self._color_mode = ColorMode.COLOR_TEMP
            self._color_temp = color_temp
            self._xy_color = None

        if xy_color is not None:
            assert self._color_cluster_handler is not None

            result = await self._color_cluster_handler.move_to_color(
                color_x=int(xy_color[0] * 65535),
                color_y=int(xy_color[1] * 65535),
                transition_time=int(10 * transition_time),
            )
            t_log["move_to_color"] = result
            if result[1] is not Status.SUCCESS:
                return False
            self._color_mode = ColorMode.XY
            self._xy_color = xy_color
            self._color_temp = None

        return True

    @property
    def is_transitioning(self) -> bool:
        """Return if the light is transitioning."""
        return self._transitioning_individual or self._transitioning_group

    def async_transition_set_flag(self) -> None:
        """Set _transitioning to True."""
        self.debug("setting transitioning flag to True")
        self._transitioning_individual = True
        self._transitioning_group = False
        if isinstance(self, LightGroup):
            for platform_entity in self.group.get_platform_entities(Light.PLATFORM):
                assert isinstance(platform_entity, Light)
                platform_entity.transition_on()

        self._async_unsub_transition_listener()

    def async_transition_start_timer(self, transition_time) -> None:
        """Start a timer to unset _transitioning_individual after transition_time.

        If necessary.
        """
        if not transition_time:
            return
        # For longer transitions, we want to extend the timer a bit more
        if transition_time >= DEFAULT_LONG_TRANSITION_TIME:
            transition_time += DEFAULT_EXTRA_TRANSITION_DELAY_LONG
        self.debug("starting transitioning timer for %s", transition_time)
        self._transition_listener = asyncio.get_running_loop().call_later(
            transition_time,
            self.async_transition_complete,
        )
        self._tracked_handles.append(self._transition_listener)

    def _async_unsub_transition_listener(self) -> None:
        """Unsubscribe transition listener."""
        if self._transition_listener:
            self._transition_listener.cancel()
            self._transition_listener = None

            with contextlib.suppress(ValueError):
                self._tracked_handles.remove(self._transition_listener)

    def async_transition_complete(self, _=None) -> None:
        """Set _transitioning_individual to False and write HA state."""
        self.debug("transition complete - future attribute reports will write HA state")
        self._transitioning_individual = False
        self._async_unsub_transition_listener()
        self.maybe_emit_state_changed_event()
        if isinstance(self, LightGroup):
            for platform_entity in self.group.get_platform_entities(Light.PLATFORM):
                assert isinstance(platform_entity, Light)
                platform_entity.transition_off()

            if self._debounced_member_refresh is not None:
                self.debug("transition complete - refreshing group member states")

                self.group.gateway.async_create_task(
                    self._debounced_member_refresh.async_call(),
                    "zha.light-refresh-debounced-member",
                )

    async def on_remove(self) -> None:
        """Clean up when entity is removed."""
        self._async_unsub_transition_listener()
        await super().on_remove()


@register_entity(OnOff.cluster_id)
class Light(BaseClusterHandlerLight, PlatformEntity):
    """Representation of a ZHA or ZLL light."""

    _attr_translation_key: str = "light"
    _REFRESH_INTERVAL = (2700, 4500)
    __polling_interval: int

    _cluster_handler_match = ClusterHandlerMatch(
        cluster_handlers=frozenset({CLUSTER_HANDLER_ON_OFF}),
        optional_cluster_handlers=frozenset(
            {CLUSTER_HANDLER_COLOR, CLUSTER_HANDLER_LEVEL}
        ),
        profile_device_types=LIGHT_PROFILE_DEVICE_TYPES,
        feature_priority=(PlatformFeatureGroup.LIGHT_OR_SWITCH_OR_SHADE, 0),
    )

    def __init__(
        self,
        cluster_handlers: list[ClusterHandler],
        endpoint: Endpoint,
        device: Device,
        **kwargs,
    ) -> None:
        """Initialize the light."""
        super().__init__(
            cluster_handlers,
            endpoint,
            device,
            **kwargs,
            legacy_discovery_unique_id=f"{endpoint.device.ieee}-{endpoint.id}",
        )
        self._on_off_cluster_handler: OnOffClusterHandler = cast(
            OnOffClusterHandler, self.cluster_handlers[CLUSTER_HANDLER_ON_OFF]
        )
        self._state: bool = bool(self._on_off_cluster_handler.on_off)
        self._level_cluster_handler: LevelControlClusterHandler | None = cast(
            LevelControlClusterHandler | None,
            self.cluster_handlers.get(CLUSTER_HANDLER_LEVEL),
        )
        self._color_cluster_handler: ColorClusterHandler | None = cast(
            ColorClusterHandler | None, self.cluster_handlers.get(CLUSTER_HANDLER_COLOR)
        )
        self._identify_cluster_handler: IdentifyClusterHandler | None = cast(
            IdentifyClusterHandler | None, device.identify_ch
        )

        self._refresh_task: asyncio.Task | None = None

        self.recompute_capabilities()

    @property
    def _gateway(self) -> Gateway:
        return self._device.gateway

    def on_add(self) -> None:
        """Run when entity is added."""
        super().on_add()
        self._on_remove_callbacks.append(
            self._on_off_cluster_handler.on_event(
                CLUSTER_HANDLER_ATTRIBUTE_UPDATED,
                self.handle_cluster_handler_attribute_updated,
            )
        )

        if self._level_cluster_handler:
            self._on_remove_callbacks.append(
                self._level_cluster_handler.on_event(
                    CLUSTER_HANDLER_LEVEL_CHANGED, self.handle_cluster_handler_set_level
                )
            )

        self.start_polling()

    def recompute_capabilities(self) -> None:
        """Recompute capabilities."""
        super().recompute_capabilities()

        effect_list = [EFFECT_OFF]

        self._internal_supported_color_modes = {ColorMode.ONOFF}
        if self._level_cluster_handler:
            self._internal_supported_color_modes.add(ColorMode.BRIGHTNESS)
            self._supported_features |= LightEntityFeature.TRANSITION
            self._brightness = self._level_cluster_handler.current_level

        if self._color_cluster_handler:
            self._min_mireds: int = self._color_cluster_handler.min_mireds
            self._max_mireds: int = self._color_cluster_handler.max_mireds

            if self._color_cluster_handler.color_temp_supported:
                self._internal_supported_color_modes.add(ColorMode.COLOR_TEMP)
                self._color_temp = self._color_cluster_handler.color_temperature

            if self._color_cluster_handler.xy_supported:
                self._internal_supported_color_modes.add(ColorMode.XY)
                curr_x = self._color_cluster_handler.current_x
                curr_y = self._color_cluster_handler.current_y
                if curr_x is not None and curr_y is not None:
                    self._xy_color = (curr_x / 65535, curr_y / 65535)
                else:
                    self._xy_color = (0, 0)

            if self._color_cluster_handler.color_loop_supported:
                self._supported_features |= LightEntityFeature.EFFECT
                effect_list.append(EFFECT_COLORLOOP)
                if self._color_cluster_handler.color_loop_active == 1:
                    self._effect = EFFECT_COLORLOOP

        self._supported_color_modes = supported_color_modes = (
            filter_supported_color_modes(self._internal_supported_color_modes)
        )
        if len(supported_color_modes) == 1:
            self._color_mode = next(iter(supported_color_modes))
        else:  # Light supports color_temp + xy, determine which mode the light is in
            assert self._color_cluster_handler
            if (
                self._color_cluster_handler.color_mode
                == Color.ColorMode.Color_temperature
            ):
                self._color_mode = ColorMode.COLOR_TEMP
            else:
                self._color_mode = ColorMode.XY

        if self._identify_cluster_handler:
            self._supported_features |= LightEntityFeature.FLASH

        self._effect_list = effect_list

    def handle_cluster_handler_set_level(self, event: LevelChangeEvent) -> None:
        """Set the brightness of this light between 0..254.

        brightness level 255 is a special value instructing the device to come
        on at `on_level` Zigbee attribute value, regardless of the last set
        level
        """
        if self.is_transitioning:
            self.debug(
                "received level change event %s while transitioning - skipping update",
                event,
            )
            return
        value = max(0, min(254, event.level))
        self._brightness = value
        self.maybe_emit_state_changed_event()

    def start_polling(self) -> None:
        """Start polling."""
        self._refresh_task = self.device.gateway.async_create_background_task(
            self._refresh(),
            name=f"light_refresh_{self.unique_id}",
            eager_start=True,
            untracked=True,
        )
        self._tracked_tasks.append(self._refresh_task)
        self.debug(
            "started polling with refresh interval of %s",
            getattr(self, "__polling_interval"),
        )

    def enable(self) -> None:
        """Enable the entity."""
        super().enable()
        self.start_polling()

    def disable(self) -> None:
        """Disable the entity."""
        super().disable()
        if self._refresh_task:
            self._tracked_tasks.remove(self._refresh_task)
            self._refresh_task.cancel()
            self._refresh_task = None

    @periodic(_REFRESH_INTERVAL)
    async def _refresh(self) -> None:
        """Call async_get_state at an interval."""
        await self.async_update()

    def transition_on(self):
        """Handle a transition start event from a group."""
        self.debug("group transition started - setting member transitioning flag")
        self._transitioning_group = True

    def transition_off(self):
        """Handle a transition finished event from a group."""
        self.debug("group transition completed - unsetting member transitioning flag")
        self._transitioning_group = False

    def handle_cluster_handler_attribute_updated(
        self, event: ClusterAttributeUpdatedEvent
    ) -> None:
        """Set the state."""
        if (
            event.cluster_id != self._on_off_cluster_handler.cluster.cluster_id
            or event.attribute_id != OnOff.AttributeDefs.on_off.id
        ):
            return
        if self.is_transitioning:
            self.debug(
                "received onoff %s while transitioning - skipping update",
                event.attribute_value,
            )
            return
        self._state = bool(event.attribute_value)
        if event.attribute_value:
            self._off_with_transition = False
            self._off_brightness = None
        self.maybe_emit_state_changed_event()

    async def async_update(self) -> None:
        """Attempt to retrieve the state from the light."""
        if self.is_transitioning:
            self.debug("skipping async_update while transitioning")
            return
        if not self.device.gateway.config.allow_polling or not self._device.available:
            self.debug(
                "skipping polling for updated state, available: %s, allow polled requests: %s",
                self._device.available,
                self.device.gateway.config.allow_polling,
            )
            return
        self.debug("polling current state")

        if self._on_off_cluster_handler:
            state = await self._on_off_cluster_handler.get_attribute_value(
                "on_off", from_cache=False
            )
            # check if transition started whilst waiting for polled state
            if self.is_transitioning:
                return  # type: ignore[unreachable]

            if state is not None:
                self._state = state
                if state:  # reset "off with transition" flag if the light is on
                    self._off_with_transition = False
                    self._off_brightness = None

        if self._level_cluster_handler:
            level = await self._level_cluster_handler.get_attribute_value(
                "current_level", from_cache=False
            )
            # check if transition started whilst waiting for polled state
            if self.is_transitioning:
                return  # type: ignore[unreachable]
            if level is not None:
                self._brightness = level

        if self._color_cluster_handler:
            attributes = [
                "color_mode",
                "current_x",
                "current_y",
            ]
            if self._color_cluster_handler.color_temp_supported:
                attributes.append("color_temperature")
            if self._color_cluster_handler.color_loop_supported:
                attributes.append("color_loop_active")

            results = await self._color_cluster_handler.get_attributes(
                attributes, from_cache=False, only_cache=False
            )

            # although rare, a transition might have been started while we were waiting
            # for the polled attributes, so abort if we are transitioning,
            # as that state will not be accurate
            if self.is_transitioning:
                return  # type: ignore[unreachable]

            if (color_mode := results.get("color_mode")) is not None:
                if color_mode == Color.ColorMode.Color_temperature:
                    self._color_mode = ColorMode.COLOR_TEMP
                    color_temp = results.get("color_temperature")
                    if color_temp is not None and color_mode:
                        self._color_temp = color_temp
                        self._xy_color = None
                else:
                    self._color_mode = ColorMode.XY
                    color_x = results.get("current_x")
                    color_y = results.get("current_y")
                    if color_x is not None and color_y is not None:
                        self._xy_color = (color_x / 65535, color_y / 65535)
                        self._color_temp = None

            color_loop_active = results.get("color_loop_active")
            if color_loop_active is not None:
                if color_loop_active == 1:
                    self._effect = EFFECT_COLORLOOP
                else:
                    self._effect = EFFECT_OFF
        self.maybe_emit_state_changed_event()


@register_entity(OnOff.cluster_id)
class HueLight(Light):
    """Representation of a HUE light which does not report attributes."""

    _REFRESH_INTERVAL = (180, 300)

    _cluster_handler_match = ClusterHandlerMatch(
        cluster_handlers=frozenset({CLUSTER_HANDLER_ON_OFF}),
        optional_cluster_handlers=frozenset(
            {CLUSTER_HANDLER_COLOR, CLUSTER_HANDLER_LEVEL}
        ),
        manufacturers=frozenset({"Philips", "Signify Netherlands B.V."}),
        profile_device_types=LIGHT_PROFILE_DEVICE_TYPES,
        # We want this entity to be preferred over the base light
        feature_priority=(PlatformFeatureGroup.LIGHT_OR_SWITCH_OR_SHADE, 1),
    )


@register_entity(OnOff.cluster_id)
class ForceOnLight(Light):
    """Representation of a light which does not respect on/off for move_to_level_with_on_off commands."""

    _FORCE_ON = True

    _cluster_handler_match = ClusterHandlerMatch(
        cluster_handlers=frozenset({CLUSTER_HANDLER_ON_OFF}),
        optional_cluster_handlers=frozenset(
            {CLUSTER_HANDLER_COLOR, CLUSTER_HANDLER_LEVEL}
        ),
        manufacturers=frozenset(
            {
                "Jasco",
                "Jasco Products",
                "Quotra-Vision",
                "eWeLight",
                "eWeLink",
            }
        ),
        profile_device_types=LIGHT_PROFILE_DEVICE_TYPES,
        # We want this entity to be preferred over the base light
        feature_priority=(PlatformFeatureGroup.LIGHT_OR_SWITCH_OR_SHADE, 1),
    )


@register_entity(OnOff.cluster_id)
class MinTransitionLight(Light):
    """Representation of a light which does not react to any "move to" calls with 0 as a transition."""

    # Transitions are counted in 1/10th of a second increments, so this is the smallest
    _DEFAULT_MIN_TRANSITION_TIME = 0.1

    _cluster_handler_match = ClusterHandlerMatch(
        cluster_handlers=frozenset({CLUSTER_HANDLER_ON_OFF}),
        optional_cluster_handlers=frozenset(
            {CLUSTER_HANDLER_COLOR, CLUSTER_HANDLER_LEVEL}
        ),
        manufacturers=DEFAULT_MIN_TRANSITION_MANUFACTURERS,
        profile_device_types=LIGHT_PROFILE_DEVICE_TYPES,
        # We want this entity to be preferred over the base light
        feature_priority=(PlatformFeatureGroup.LIGHT_OR_SWITCH_OR_SHADE, 1),
    )


@register_group_entity
class LightGroup(BaseClusterHandlerLight, GroupEntity):
    """Representation of a light group."""

    _attr_always_supported = True

    def __init__(self, group: Group):
        """Initialize a light group."""
        super().__init__(group)

        # TODO: these type annotations are not correct, these are not cluster handler
        # objects but are instead cluster proxies!
        self._on_off_cluster_handler: OnOffClusterHandler = cast(
            OnOffClusterHandler, group.zigpy_group.endpoint[OnOff.cluster_id]
        )
        self._level_cluster_handler: LevelControlClusterHandler | None = cast(
            LevelControlClusterHandler | None,
            group.zigpy_group.endpoint[LevelControl.cluster_id],
        )
        self._color_cluster_handler: ColorClusterHandler | None = cast(
            ColorClusterHandler | None, group.zigpy_group.endpoint[Color.cluster_id]
        )
        self._identify_cluster_handler: IdentifyClusterHandler | None = cast(
            IdentifyClusterHandler | None,
            group.zigpy_group.endpoint[Identify.cluster_id],
        )

        self._debounced_member_refresh: Debouncer | None = Debouncer(
            self.group.gateway,
            _LOGGER,
            cooldown=3,
            immediate=True,
            function=self._force_member_updates,
        )

        self.recompute_capabilities()

    @property
    def _gateway(self) -> Gateway:
        return self.group.gateway

    def recompute_capabilities(self) -> None:
        """Recompute capabilities."""
        super().recompute_capabilities()

        light_options = self._gateway.config.config.light_options
        self._zha_config_group_members_assume_state = (
            light_options.group_members_assume_state
        )

        self._zha_config_enhanced_light_transition = False
        self._GROUP_SUPPORTS_EXECUTE_IF_OFF: bool = True

        if light_options.group_members_assume_state:
            self._change_listener_debouncer.cooldown = (
                ASSUME_UPDATE_GROUP_FROM_CHILD_DELAY
            )
        else:
            self._change_listener_debouncer.cooldown = (
                DEFAULT_UPDATE_GROUP_FROM_CHILD_DELAY
            )

        for member in self.group.members:
            # Ensure we do not send group commands that violate the minimum transition
            # time of any members.
            if member.device.manufacturer in DEFAULT_MIN_TRANSITION_MANUFACTURERS:
                self._DEFAULT_MIN_TRANSITION_TIME = (
                    MinTransitionLight._DEFAULT_MIN_TRANSITION_TIME
                )

            # Check all group members to see if they support execute_if_off.
            # If at least one member has a color cluster and doesn't support it,
            # it's not used.
            for endpoint in member.device._endpoints.values():
                for cluster_handler in endpoint.all_cluster_handlers.values():
                    if (
                        cluster_handler.name == CLUSTER_HANDLER_COLOR
                        and not cluster_handler.execute_if_off_supported
                    ):
                        self._GROUP_SUPPORTS_EXECUTE_IF_OFF = False
                        break

        self._color_mode = ColorMode.UNKNOWN
        self._internal_supported_color_modes = {ColorMode.ONOFF}

        if hasattr(self, "info_object"):
            delattr(self, "info_object")
        self.update()

    async def on_remove(self) -> None:
        """Cancel tasks this entity owns."""
        await super().on_remove()
        if self._debounced_member_refresh:
            self._debounced_member_refresh.async_cancel()

    async def async_turn_on(self, **kwargs: Any) -> None:
        """Turn the entity on."""
        # "off with transition" and "off brightness" will get overridden when
        # turning on the group, but they are needed for setting the assumed
        # member state correctly, so save them here
        off_brightness = self._off_brightness if self._off_with_transition else None
        await super().async_turn_on(**kwargs)
        if self._zha_config_group_members_assume_state:
            self._make_members_assume_group_state(
                state=True, off_brightness=off_brightness, **kwargs
            )
        if self.is_transitioning:  # when transitioning, state is refreshed at the end
            return
        if self._debounced_member_refresh:
            await self._debounced_member_refresh.async_call()

    async def async_turn_off(self, **kwargs: Any) -> None:
        """Turn the entity off."""
        await super().async_turn_off(**kwargs)
        if self._zha_config_group_members_assume_state:
            self._make_members_assume_group_state(state=False, **kwargs)
        if self.is_transitioning:
            return
        if self._debounced_member_refresh:
            await self._debounced_member_refresh.async_call()

    def update(self, _: Any = None) -> None:
        """Query all members and determine the light group state."""
        self.debug("Updating light group entity state")
        platform_entities = self._group.get_platform_entities(self.PLATFORM)
        all_states = [entity.state for entity in platform_entities]
        states: list = list(filter(None, all_states))
        self.debug(
            "All platform entity states for group entity members: %s", all_states
        )
        on_states = [state for state in states if state["on"]]

        self._state = len(on_states) > 0

        # reset "off with transition" flag if any member is on
        if self._state:
            self._off_with_transition = False
            self._off_brightness = None

        self._brightness = reduce_attribute(on_states, ATTR_BRIGHTNESS)
        self._xy_color = reduce_attribute(on_states, ATTR_XY_COLOR, reduce=mean_tuple)
        self._color_temp = reduce_attribute(on_states, ATTR_COLOR_TEMP)
        self._min_mireds = reduce_attribute(
            states, ATTR_MIN_MIREDS, default=153, reduce=min
        )
        self._max_mireds = reduce_attribute(
            states, ATTR_MAX_MIREDS, default=500, reduce=max
        )

        self._effect_list = None
        all_effect_lists = list(find_state_attributes(states, ATTR_EFFECT_LIST))
        if all_effect_lists:
            # Merge all effects from all effect_lists with a union merge.
            self._effect_list = list(set().union(*all_effect_lists))

        self._effect = EFFECT_OFF
        all_effects = list(find_state_attributes(on_states, ATTR_EFFECT))
        if all_effects:
            # Report the most common effect.
            effects_count = Counter(itertools.chain(all_effects))
            self._effect = effects_count.most_common(1)[0][0]

        supported_color_modes = {ColorMode.ONOFF}
        all_supported_color_modes: list[set[ColorMode]] = list(
            find_state_attributes(states, ATTR_SUPPORTED_COLOR_MODES)
        )
        self._internal_supported_color_modes = set().union(*all_supported_color_modes)

        if all_supported_color_modes:
            # Merge all color modes.
            self._supported_color_modes = supported_color_modes = (
                filter_supported_color_modes(set().union(*all_supported_color_modes))
            )

        self._color_mode = ColorMode.UNKNOWN
        all_color_modes = list(find_state_attributes(on_states, ATTR_COLOR_MODE))
        if all_color_modes:
            # Report the most common color mode, select brightness and onoff last
            color_mode_count = Counter(itertools.chain(all_color_modes))
            if ColorMode.ONOFF in color_mode_count:
                if ColorMode.ONOFF in supported_color_modes:
                    color_mode_count[ColorMode.ONOFF] = -1
                else:
                    color_mode_count.pop(ColorMode.ONOFF)
            if ColorMode.BRIGHTNESS in color_mode_count:
                if ColorMode.BRIGHTNESS in supported_color_modes:
                    color_mode_count[ColorMode.BRIGHTNESS] = 0
                else:
                    color_mode_count.pop(ColorMode.BRIGHTNESS)
            if color_mode_count:
                self._color_mode = color_mode_count.most_common(1)[0][0]
            else:
                self._color_mode = next(iter(supported_color_modes))

        self._supported_features = LightEntityFeature(0)
        for support in find_state_attributes(states, ATTR_SUPPORTED_FEATURES):
            # Merge supported features by emulating support for every feature
            # we find.
            self._supported_features |= support
        # Bitwise-and the supported features with the GroupedLight's features
        # so that we don't break in the future when a new feature is added.
        self._supported_features &= (
            LightEntityFeature.EFFECT
            | LightEntityFeature.FLASH
            | LightEntityFeature.TRANSITION
        )
        self.maybe_emit_state_changed_event()

    async def _force_member_updates(self) -> None:
        """Force the update of members to ensure the states are correct for bulbs that don't report their state."""
        for platform_entity in self.group.get_platform_entities(Light.PLATFORM):
            await platform_entity.async_update()

    def _make_members_assume_group_state(
        self,
        *,
        state: bool,
        off_brightness: int | None = None,
        # The rest are service call kwargs
        transition: float | None = None,
        brightness: int | None = None,
        effect: str | None = None,
        flash: FlashMode | None = None,
        color_temp: int | None = None,
        xy_color: tuple[int, int] | None = None,
    ) -> None:
        """Send an assume event to all members of the group."""

        # check if the parameters were actually updated
        # in the service call before updating members
        if brightness is not None:  # or off brightness
            assumed_brightness = self._brightness
        elif off_brightness is not None:
            # if we turn on the group light with "off brightness",
            # pass that to the members
            assumed_brightness = off_brightness
        else:
            assumed_brightness = None

        if color_temp is not None:
            assumed_color_mode = self._color_mode
            assumed_color_temp = self._color_temp
        else:
            assumed_color_mode = None
            assumed_color_temp = None

        if xy_color is not None:
            assumed_color_mode = self._color_mode
            assumed_xy_color = self._xy_color
        else:
            assumed_color_mode = None
            assumed_xy_color = None

        # setting any other attribute will turn the effect off, so we always set this
        assumed_effect = self._effect

        for platform_entity in self.group.get_platform_entities(Light.PLATFORM):
            assert isinstance(platform_entity, Light)
            if not platform_entity.available:
                continue

            supported_modes = platform_entity._internal_supported_color_modes
            effect_list = platform_entity._effect_list

            brightness_supported = (
                is_brightness_supported(supported_modes)
                and platform_entity._level_cluster_handler is not None
            )

            # unset "off brightness" and "off with transition"
            # if group turned on this light
            if state and not platform_entity._state:
                platform_entity._off_with_transition = False
                platform_entity._off_brightness = None

            # set "off brightness" and "off with transition"
            # if group turned off this light, and the light was not already off
            # (to not override _off_with_transition)
            elif not state and platform_entity._state and brightness_supported:
                # use individual brightness, instead of possibly averaged
                # brightness from group
                platform_entity._off_brightness = platform_entity._brightness
                platform_entity._off_with_transition = self._off_with_transition

            # Note: If individual lights have off_with_transition set, but not the
            # group, and the group is then turned on without a level, individual lights
            # might fall back to brightness level 1.
            # Since all lights might need different brightness levels to be turned on,
            # we can't use one group call. And making individual calls when turning on
            # a ZHA group would cause a lot of traffic. In this case,
            # turn_on should either just be called with a level or individual turn_on
            # calls can be used.

            # state is always set (turn_on/turn_off)
            platform_entity._state = state

            # before assuming a group state attribute, check if the attribute
            # was actually set in that call
            if assumed_brightness is not None and brightness_supported:
                platform_entity._brightness = assumed_brightness
            if assumed_color_mode is not None and assumed_color_mode in supported_modes:
                platform_entity._color_mode = assumed_color_mode
            if (
                assumed_color_temp is not None
                and ColorMode.COLOR_TEMP in supported_modes
            ):
                platform_entity._color_temp = assumed_color_temp
            if assumed_xy_color is not None and ColorMode.XY in supported_modes:
                platform_entity._xy_color = assumed_xy_color
            if (
                assumed_effect is not None
                and effect_list
                and assumed_effect in effect_list
            ):
                platform_entity._effect = assumed_effect

            platform_entity.maybe_emit_state_changed_event()

    def restore_external_state_attributes(
        self,
        *,
        state: bool | None,
        off_with_transition: bool | None,
        off_brightness: int | None,
        brightness: int | None,
        color_temp: int | None,
        xy_color: tuple[float, float] | None,
        color_mode: ColorMode | None,
        effect: str | None,
    ) -> None:
        """Restore extra state attributes."""

        # Group state is calculated from the members,
        # except for off_with_transition and off_brightness
        if off_with_transition is not None:
            self._off_with_transition = off_with_transition
        if off_brightness is not None:
            self._off_brightness = off_brightness
